#!/usr/bin/env python3

"""
Functional tests for 'align' subcommand
"""
import os
import shutil
import subprocess
import pytest
import logging

import PanACoTA.subcommands.align as al
import test.test_unit.utilities_for_tests as tutil
# from PanACoTA import utils


# Define common variables
ALDIR = os.path.join("test", "data", "align")
EXPPATH = os.path.join(ALDIR, "exp_files")
TESTPATH = os.path.join(ALDIR, "test_files")
LOGFILE_BASE = "logs_test_postalign"
GENEPATH = os.path.join(ALDIR, "generated_by_func_tests")

@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    if not os.path.isdir(GENEPATH):
        os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")


def test_main():
    """
    Test that when giving a database, a persistent genome and a list of genomes, it extracts
    expected proteins by family, aligns each family, back-translates them, concatenates all
    families into one file and groups them by genome.
    """
    corepers = os.path.join(TESTPATH, "test_pers0.99FX.lst")
    list_genomes = os.path.join("test", "data", "pangenome", "test_files", "list_to_pan.txt")
    dname = "TEST4"
    dbpath = os.path.join("test", "data", "pangenome", "test_files", "example_db")
    outdir = GENEPATH
    threads = 1
    force = False
    cmd = "cmd"
    al.main(cmd, corepers, list_genomes, dname, dbpath, outdir, threads, force)
    # Check creation of the 3 subdirectories
    aldir = os.path.join(outdir, "Align-" + dname)
    listdir = os.path.join(outdir, "List-" + dname)
    treedir = os.path.join(outdir, "Phylo-" + dname)
    assert os.path.isdir(aldir)
    assert os.path.isdir(listdir)
    assert os.path.isdir(treedir)
    # Check content of listdir
    genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    for gen in genomes:
        assert os.path.isfile(os.path.join(listdir, f"{dname}-getEntry_gen_{gen}.txt"))
        assert os.path.isfile(os.path.join(listdir, f"{dname}-getEntry_prt_{gen}.txt"))
    # Check content of aldir
    fams = [1, 4, 6, 8, 10, 11, 13, 14]
    for fam in fams:
        assert os.path.isfile(os.path.join(aldir, f'{dname}-current.{fam}.gen'))
        assert os.path.isfile(os.path.join(aldir, f'{dname}-current.{fam}.prt'))
        assert os.path.isfile(os.path.join(aldir, f'{dname}-current.{fam}.miss.lst'))
        assert os.path.isfile(os.path.join(aldir, f'{dname}-mafft-align.{fam}.aln'))
        assert os.path.isfile(os.path.join(aldir, f'{dname}-mafft-prt2nuc.{fam}.aln'))
    out_concat = os.path.join(aldir, dname + "-complete.cat.aln")
    exp_concat = os.path.join(EXPPATH, "exp_pers4genome-complete.cat.aln")
    assert tutil.compare_order_content(out_concat, exp_concat)
    # Check content of treedir
    out_grp = os.path.join(treedir, dname + ".grp.aln")
    exp_grp = os.path.join(EXPPATH, "exp_pers4genomes.grp.aln")
    assert tutil.compare_order_content(out_grp, exp_grp)
    # Check presence of log files, and log.err is empty
    base_log = os.path.join(outdir, "PanACoTA-align_" + dname + ".log")
    assert os.path.isfile(base_log)
    assert os.path.isfile(base_log + ".details")
    assert os.path.isfile(base_log + ".err")
    with open(base_log + ".err", "r") as bf:
        assert bf.readlines() == []
    # Check logs
    with open(base_log + ".details", "r") as lc:
        log_content = lc.readlines()
    assert ("Reading PersGenome and constructing lists of missing genomes in "
            "each family") in " ".join(log_content)
    assert "Extracting proteins and genes from all genomes" in " ".join(log_content)
    for gen in genomes:
        assert "Extracting proteins and genes from {}".format(gen) in " ".join(log_content)
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in " ".join(log_content)
    for fam in fams:
        assert "Checking extractions for family {}".format(fam) in " ".join(log_content)
        assert "Aligning family {}".format(fam) in " ".join(log_content)
        assert "Back-translating family {}".format(fam) in " ".join(log_content)
    assert "Concatenating all alignment files" in " ".join(log_content)
    assert "Grouping alignments per genome" in " ".join(log_content)
    assert "END" in " ".join(log_content)


def test_main_exist_ok():
    """
    Test main all files exist and are ok, no force -> end without error, with warnings on re-use
    """
    corepers = os.path.join(TESTPATH, "test_pers0.99FX.lst")
    list_genomes = os.path.join("test", "data", "pangenome", "test_files", "list_to_pan.txt")
    dname = "TEST4exists"
    dbpath = os.path.join("test", "data", "pangenome", "test_files", "example_db")
    outdir = os.path.join(GENEPATH, "test_main_exist_ok")
    threads = 1
    force = False
    cmd = "cmd test_main_exist_ok"
    # Create output directories and files
    aldir = os.path.join(outdir, "Align-" + dname)
    listdir = os.path.join(outdir, "List-" + dname)
    treedir = os.path.join(outdir, "Phylo-" + dname)
    os.makedirs(aldir)
    os.makedirs(listdir)
    os.makedirs(treedir)
    # Create content of listdir
    ex_listdir = os.path.join(EXPPATH, "exp_listdir-pers")
    genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    for gen in genomes:
        outgen = os.path.join(listdir, f"{dname}-getEntry_gen_{gen}.txt")
        refgen = os.path.join(ex_listdir, f"getEntry_gen_{gen}")
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(listdir, f"{dname}-getEntry_prt_{gen}.txt")
        refprt = os.path.join(ex_listdir, f"getEntry_prt_{gen}")
        shutil.copyfile(refprt, outprt)
    # Create content of aldir
    ex_aldir = os.path.join(EXPPATH, "exp_aldir-pers")
    fams = [1, 4, 6, 8, 10, 11, 13, 14]
    for fam in fams:
        outgen = os.path.join(aldir, f'{dname}-current.{fam}.gen')
        refgen = os.path.join(ex_aldir, f"current.{fam}.gen")
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(aldir, f'{dname}-current.{fam}.prt')
        refprt = os.path.join(ex_aldir, f"current.{fam}.prt")
        shutil.copyfile(refprt, outprt)
        outmiss = os.path.join(aldir, f'{dname}-current.{fam}.miss.lst')
        refmiss = os.path.join(ex_aldir, f"current.{fam}.miss.lst")
        shutil.copyfile(refmiss, outmiss)
        outaln = os.path.join(aldir, f'{dname}-mafft-align.{fam}.aln')
        refaln = os.path.join(ex_aldir, f"mafft-align.{fam}.aln")
        shutil.copyfile(refaln, outaln)
        outbtr = os.path.join(aldir, f'{dname}-mafft-prt2nuc.{fam}.aln')
        refbtr = os.path.join(ex_aldir, f"mafft-prt2nuc.{fam}.aln")
        shutil.copyfile(refbtr, outbtr)
    outcat = os.path.join(aldir, dname + "-complete.cat.aln")
    refcat = os.path.join(EXPPATH, "exp_pers4genome-complete.cat.aln")
    shutil.copyfile(refcat, outcat)
    # Create content of treedir
    outgrp = os.path.join(treedir, dname + ".grp.aln")
    refgrp = os.path.join(EXPPATH, "exp_pers4genomes.grp.aln")
    shutil.copyfile(refgrp, outgrp)
    # Run align module
    al.main(cmd, corepers, list_genomes, dname, dbpath, outdir, threads, force)
    # Check logs
    logfile = os.path.join(outdir, "PanACoTA-align_TEST4exists.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Reading PersGenome and constructing lists of missing genomes in "
            "each family") in " ".join(log_content)
    for gen in genomes:
        assert (f"For genome {gen}, test/data/align/generated_by_func_tests/test_main_exist_ok/"
                f"List-TEST4exists/TEST4exists-getEntry_prt_{gen}.txt and test/data/align/"
                "generated_by_func_tests/test_main_exist_ok/List-TEST4exists/"
                f"TEST4exists-getEntry_gen_{gen}.txt already exist. The program "
                "will use them to extract proteins and genes. If you prefer to rewrite "
                "them, use option -F ") in " ".join(log_content)
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in " ".join(log_content)
    for fam in fams:
        assert "Checking extractions for family {}".format(fam) in " ".join(log_content)
        assert ("Alignment already done for family {}. The program will use it for next "
                "steps").format(fam) in " ".join(log_content)
    assert ("All extraction files already existing (see detailed log for more "
            "information)") in " ".join(log_content)
    assert ("All prt and gene files for all families already exist. The program will use them "
            "for the next step. If you want to re-extract a given family, remove its prt and "
            "gen extraction files. If you want to re-extract all families, use option -F "
            "(or --force).") in " ".join(log_content)
    assert ("Alignments already concatenated in "
            "test/data/align/generated_by_func_tests/test_main_exist_ok/Align-TEST4exists/"
            "TEST4exists-complete.cat.aln. Program will "
            "use it for next steps. If you want to redo it, remove it before "
            "running.") in " ".join(log_content)
    assert ("Alignments already grouped by genome in "
            "test/data/align/generated_by_func_tests/test_main_exist_ok/Phylo-TEST4exists/"
            "TEST4exists.grp.aln. Program will "
            "end.") in " ".join(log_content)
    assert "END" in " ".join(log_content)


def test_main_exist_emptygrp(capsys):
    """
    test main all files exist but empty grp -> does nothing, grp is not checked if everything
    before was ok. grp must still be empty
    """
    corepers = os.path.join(TESTPATH, "test_pers0.99FX.lst")
    list_genomes = os.path.join("test", "data", "pangenome", "test_files", "list_to_pan.txt")
    dname = "TEST4empty-grp"
    dbpath = os.path.join("test", "data", "pangenome", "test_files", "example_db")
    outdir = os.path.join(GENEPATH, "test_main_exist_ok")
    threads = 1
    force = False
    cmd = "cmd test_main_exist_emptygrp"
    # Create output directories and files
    aldir = os.path.join(outdir, "Align-" + dname)
    listdir = os.path.join(outdir, "List-" + dname)
    treedir = os.path.join(outdir, "Phylo-" + dname)
    os.makedirs(aldir)
    os.makedirs(listdir)
    os.makedirs(treedir)
    # Create content of listdir
    ex_listdir = os.path.join(EXPPATH, "exp_listdir-pers")
    genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    for gen in genomes:
        outgen = os.path.join(listdir, "{}-getEntry_gen_{}.txt".format(dname, gen))
        refgen = os.path.join(ex_listdir, "getEntry_gen_{}".format(gen))
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(listdir, "{}-getEntry_prt_{}.txt".format(dname, gen))
        refprt = os.path.join(ex_listdir, "getEntry_prt_{}".format(gen))
        shutil.copyfile(refprt, outprt)
    # Create content of aldir
    ex_aldir = os.path.join(EXPPATH, "exp_aldir-pers")
    fams = [1, 4, 6, 8, 10, 11, 13, 14]
    for fam in fams:
        outgen = os.path.join(aldir, '{}-current.{}.gen').format(dname, fam)
        refgen = os.path.join(ex_aldir, "current.{}.gen").format(fam)
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(aldir, '{}-current.{}.prt').format(dname, fam)
        refprt = os.path.join(ex_aldir, "current.{}.prt").format(fam)
        shutil.copyfile(refprt, outprt)
        outmiss = os.path.join(aldir, '{}-current.{}.miss.lst').format(dname, fam)
        refmiss = os.path.join(ex_aldir, "current.{}.miss.lst").format(fam)
        shutil.copyfile(refmiss, outmiss)
        outaln = os.path.join(aldir, '{}-mafft-align.{}.aln').format(dname, fam)
        refaln = os.path.join(ex_aldir, "mafft-align.{}.aln").format(fam)
        shutil.copyfile(refaln, outaln)
        outbtr = os.path.join(aldir, '{}-mafft-prt2nuc.{}.aln').format(dname, fam)
        refbtr = os.path.join(ex_aldir, "mafft-prt2nuc.{}.aln").format(fam)
        shutil.copyfile(refbtr, outbtr)
    outcat = os.path.join(aldir, dname + "-complete.cat.aln")
    refcat = os.path.join(EXPPATH, "exp_pers4genome-complete.cat.aln")
    shutil.copyfile(refcat, outcat)
    # Create content of treedir, with empty grp
    outgrp = os.path.join(treedir, dname + ".grp.aln")
    open(outgrp, "w").close()
    # Run align module
    al.main(cmd, corepers, list_genomes, dname, dbpath, outdir, threads, force, verbose=2)
    # Check grp still empty
    assert os.stat(outgrp).st_size == 0
    # Check logs
    out, err = capsys.readouterr()
    logfile = os.path.join(outdir, "PanACoTA-align_TEST4empty-grp.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Reading PersGenome and constructing lists of missing genomes in "
            "each family") in out
    for gen in genomes:
        assert ("For genome {0}, "
                "{2}/List-{1}/{1}-getEntry_prt_{0}.txt and {2}/List-{1}/{1}-getEntry_gen_{0}.txt "
                "already exist. The program will use them to extract proteins and genes. If you "
                "prefer to rewrite them, use option -F "
                "(or --force)".format(gen, dname, outdir)) in " ".join(log_content)
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in out
    for fam in fams:
        assert "Checking extractions for family {}".format(fam) in out
        assert ("Alignment already done for family {}. The program will use it for next "
                "steps").format(fam) in " ".join(log_content)
    assert ("All extraction files already existing (see detailed log for more "
            "information)") in " ".join(log_content)
    assert ("All prt and gene files for all families already exist. The program will use them "
            "for the next step. If you want to re-extract a given family, remove its prt and "
            "gen extraction files. If you want to re-extract all families, use option -F "
            "(or --force).") in " ".join(log_content)
    assert "Alignments already concatenated" in out
    assert ("Alignments already concatenated in {1}/Align-{0}/{0}-complete.cat.aln. Program will "
            "use it for next steps. If you want to redo it, remove it before "
            "running.").format(dname, outdir) in " ".join(log_content)
    assert "Alignments already grouped by genome" in " ".join(log_content)
    assert ("Alignments already grouped by genome in "
            "test/data/align/generated_by_func_tests/test_main_exist_ok/"
            "Phylo-{0}/{0}.grp.aln. Program will end.").format(dname) in " ".join(log_content)
    assert "END" in out


def test_main_exist_emptycat(capsys):
    """
    test main all files exist but "1 sentence" concat -> do nothing, as we do not check concat file
    """
    corepers = os.path.join(TESTPATH, "test_pers0.99FX.lst")
    list_genomes = os.path.join("test", "data", "pangenome", "test_files", "list_to_pan.txt")
    dname = "TEST4empty-cat"
    dbpath = os.path.join("test", "data", "pangenome", "test_files", "example_db")
    outdir = os.path.join(GENEPATH, "test_main_exist_emptycat")
    threads = 1
    force = False
    cmd = "cmd test_main_exist_emptycat"
    # Create output directories and files
    aldir = os.path.join(outdir, "Align-" + dname)
    listdir = os.path.join(outdir, "List-" + dname)
    treedir = os.path.join(outdir, "Phylo-" + dname)
    os.makedirs(aldir)
    os.makedirs(listdir)
    os.makedirs(treedir)
    # Create content of listdir
    ex_listdir = os.path.join(EXPPATH, "exp_listdir-pers")
    genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    for gen in genomes:
        outgen = os.path.join(listdir, "{}-getEntry_gen_{}.txt".format(dname, gen))
        refgen = os.path.join(ex_listdir, "getEntry_gen_{}".format(gen))
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(listdir, "{}-getEntry_prt_{}.txt".format(dname, gen))
        refprt = os.path.join(ex_listdir, "getEntry_prt_{}".format(gen))
        shutil.copyfile(refprt, outprt)
    # Create content of aldir
    ex_aldir = os.path.join(EXPPATH, "exp_aldir-pers")
    fams = [1, 4, 6, 8, 10, 11, 13, 14]
    for fam in fams:
        outgen = os.path.join(aldir, '{}-current.{}.gen').format(dname, fam)
        refgen = os.path.join(ex_aldir, "current.{}.gen").format(fam)
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(aldir, '{}-current.{}.prt').format(dname, fam)
        refprt = os.path.join(ex_aldir, "current.{}.prt").format(fam)
        shutil.copyfile(refprt, outprt)
        outmiss = os.path.join(aldir, '{}-current.{}.miss.lst').format(dname, fam)
        refmiss = os.path.join(ex_aldir, "current.{}.miss.lst").format(fam)
        shutil.copyfile(refmiss, outmiss)
        outaln = os.path.join(aldir, '{}-mafft-align.{}.aln').format(dname, fam)
        refaln = os.path.join(ex_aldir, "mafft-align.{}.aln").format(fam)
        shutil.copyfile(refaln, outaln)
        outbtr = os.path.join(aldir, '{}-mafft-prt2nuc.{}.aln').format(dname, fam)
        refbtr = os.path.join(ex_aldir, "mafft-prt2nuc.{}.aln").format(fam)
        shutil.copyfile(refbtr, outbtr)
    outcat = os.path.join(aldir, dname + "-complete.cat.aln")
    with open(outcat, "w") as outf:
        outf.write("Hello !!")
    # Create content of treedir
    outgrp = os.path.join(treedir, dname + ".grp.aln")
    with open(outgrp, "w") as outf:
        outf.write("It's me !!")
    # Run align module
    al.main(cmd, corepers, list_genomes, dname, dbpath, outdir, threads, force, verbose=16)
    # Check concat and grp did not change
    with open(outcat, "r") as of:
        assert of.readlines() == ["Hello !!"]
    with open(outgrp, "r") as of:
        assert of.readlines() == ["It's me !!"]
    # Check logs
    out, err = capsys.readouterr()
    logfile = os.path.join(outdir, "PanACoTA-align_TEST4empty-cat.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Reading PersGenome and constructing lists of missing genomes in "
            "each family") in out
    for gen in genomes:
        assert ("For genome {0}, "
                "{2}/List-{1}/{1}-getEntry_prt_{0}.txt and {2}/List-{1}/{1}-getEntry_gen_{0}.txt "
                "already exist. The program will use them to extract proteins and genes. If you "
                "prefer to rewrite them, use option -F "
                "(or --force)".format(gen, dname, outdir)) in " ".join(log_content)
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in out
    for fam in fams:
        assert "Checking extractions for family {}".format(fam) in out
        assert ("Alignment already done for family {}. The program will use it for next "
                "steps").format(fam) in " ".join(log_content)
    assert ("All extraction files already existing (see detailed log for more "
            "information)") in " ".join(log_content)
    assert ("All prt and gene files for all families already exist. The program will use them "
            "for the next step. If you want to re-extract a given family, remove its prt and "
            "gen extraction files. If you want to re-extract all families, use option -F "
            "(or --force).") in " ".join(log_content)
    assert "Alignments already concatenated" in out
    assert ("Alignments already concatenated in {1}/Align-{0}/{0}-complete.cat.aln. Program will "
            "use it for next steps. If you want to redo it, remove it before "
            "running.".format(dname, outdir)) in " ".join(log_content)
    assert "Alignments already grouped by genome" in out
    assert ("Alignments already grouped by genome in "
            "{1}/Phylo-{0}/{0}.grp.aln. Program will "
            "end.".format(dname, outdir)) in " ".join(log_content)
    assert "END" in out


def test_main_exist_emptyp2n(capsys):
    """
    test main all files exist but empty prt2nuc for 1 fam -> redo this prt2nuc and concat and grp
    """
    corepers = os.path.join(TESTPATH, "test_pers0.99FX.lst")
    list_genomes = os.path.join("test", "data", "pangenome", "test_files", "list_to_pan.txt")
    dname = "TEST4empty-p2n"
    dbpath = os.path.join("test", "data", "pangenome", "test_files", "example_db")
    outdir = os.path.join(GENEPATH, "test_main_exist_emptyp2n")
    threads = 1
    force = False
    cmd = "cmd test_main_exist_emptycat"
    # Create output directories and files
    aldir = os.path.join(outdir, "Align-" + dname)
    listdir = os.path.join(outdir, "List-" + dname)
    treedir = os.path.join(outdir, "Phylo-" + dname)
    os.makedirs(aldir)
    os.makedirs(listdir)
    os.makedirs(treedir)
    # Create content of listdir
    ex_listdir = os.path.join(EXPPATH, "exp_listdir-pers")
    genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    for gen in genomes:
        outgen = os.path.join(listdir, "{}-getEntry_gen_{}.txt".format(dname, gen))
        refgen = os.path.join(ex_listdir, "getEntry_gen_{}".format(gen))
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(listdir, "{}-getEntry_prt_{}.txt".format(dname, gen))
        refprt = os.path.join(ex_listdir, "getEntry_prt_{}".format(gen))
        shutil.copyfile(refprt, outprt)
    # Create content of aldir
    ex_aldir = os.path.join(EXPPATH, "exp_aldir-pers")
    fams = [1, 4, 6, 8, 10, 11, 13, 14]
    for fam in fams:
        outgen = os.path.join(aldir, '{}-current.{}.gen').format(dname, fam)
        refgen = os.path.join(ex_aldir, "current.{}.gen").format(fam)
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(aldir, '{}-current.{}.prt').format(dname, fam)
        refprt = os.path.join(ex_aldir, "current.{}.prt").format(fam)
        shutil.copyfile(refprt, outprt)
        outmiss = os.path.join(aldir, '{}-current.{}.miss.lst').format(dname, fam)
        refmiss = os.path.join(ex_aldir, "current.{}.miss.lst").format(fam)
        shutil.copyfile(refmiss, outmiss)
        outaln = os.path.join(aldir, '{}-mafft-align.{}.aln').format(dname, fam)
        refaln = os.path.join(ex_aldir, "mafft-align.{}.aln").format(fam)
        shutil.copyfile(refaln, outaln)
        outbtr = os.path.join(aldir, '{}-mafft-prt2nuc.{}.aln').format(dname, fam)
        refbtr = os.path.join(ex_aldir, "mafft-prt2nuc.{}.aln").format(fam)
        shutil.copyfile(refbtr, outbtr)
    outbtr1 = os.path.join(aldir, '{}-mafft-prt2nuc.1.aln').format(dname)
    with open(outbtr1, "w") as outf:
        outf.write("Hello !!")
    outcat = os.path.join(aldir, dname + "-complete.cat.aln")
    with open(outcat, "w") as oc:
        oc.write("complete cat aln")
    # Create content of treedir
    outgrp = os.path.join(treedir, dname + ".grp.aln")
    with open(outgrp, "w") as og:
        og.write("grouped grp aln")
    # Run align module
    al.main(cmd, corepers, list_genomes, dname, dbpath, outdir, threads, force)
    # Check btr file for fam 1
    ref_btr1 = os.path.join(ex_aldir, "mafft-prt2nuc.1.aln")
    assert tutil.compare_order_content(ref_btr1, outbtr1)
    # Check concat and group files
    refcat = os.path.join(ex_aldir, "complete.cat.aln")
    assert tutil.compare_order_content(refcat, outcat)
    refgrp = os.path.join(EXPPATH, "exp_pers4genomes.grp.aln")
    assert tutil.compare_order_content(refgrp, outgrp)
    # Check logs
    out, err = capsys.readouterr()
    logfile = os.path.join(outdir, "PanACoTA-align_TEST4empty-p2n.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Reading PersGenome and constructing lists of missing genomes in "
            "each family") in out
    for gen in genomes:
        assert ("For genome {0}, "
                "{2}/List-{1}/{1}-getEntry_prt_{0}.txt and {2}/List-{1}/{1}-getEntry_gen_{0}.txt "
                "already exist. The program will use them to extract proteins and genes. If you "
                "prefer to rewrite them, use option -F "
                "(or --force)".format(gen, dname, outdir)) in " ".join(log_content)
    assert ("All extraction files already existing (see detailed log for more "
            "information)") in out
    assert ("All prt and gene files for all families already exist. The program will use them "
            "for the next step. If you want to re-extract a given family, remove its prt and "
            "gen extraction files. If you want to re-extract all families, use option -F "
            "(or --force).") in " ".join(log_content)
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in out
    assert "Checking extractions for family 1" in " ".join(log_content)
    assert ("fam 1: Will redo back-translation, because found a different number of proteins "
            "aligned in {0}/Align-{1}/{1}-mafft-align.1.aln (4) and genes back-translated in "
            "existing {0}/Align-{1}/"
            "{1}-mafft-prt2nuc.1.aln".format(outdir, dname)) in " ".join(log_content)
    assert "Back-translating family 1" in " ".join(log_content)
    fams2 = fams[1:]
    for fam in fams2:
        assert "Checking extractions for family {}".format(fam) in " ".join(log_content)
        assert ("Alignment already done for family {}. The program will use it for next "
                "steps").format(fam) in " ".join(log_content)
    assert "Concatenating all alignment files" in out
    assert "Grouping alignments per genome" in out
    assert "END" in out


def test_main_emptymafft1(capsys):
    """
    Test main all files exist but mafft empty for 1 fam -> redo mafft and prt2nuc for this fam,
    + concat + grp
    """
    corepers = os.path.join(TESTPATH, "test_pers0.99FX.lst")
    list_genomes = os.path.join("test", "data", "pangenome", "test_files", "list_to_pan.txt")
    dname = "TEST4-emptym1"
    dbpath = os.path.join("test", "data", "pangenome", "test_files", "example_db")
    outdir = os.path.join(GENEPATH, "test_main_exist_emptym1")
    threads = 1
    force = False
    cmd = "cmd test_main_exist_empty-m1"
    # Create output directories and files
    aldir = os.path.join(outdir, "Align-" + dname)
    listdir = os.path.join(outdir, "List-" + dname)
    treedir = os.path.join(outdir, "Phylo-" + dname)
    os.makedirs(aldir)
    os.makedirs(listdir)
    os.makedirs(treedir)
    # Create content of listdir
    ex_listdir = os.path.join(EXPPATH, "exp_listdir-pers")
    genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    for gen in genomes:
        outgen = os.path.join(listdir, "{}-getEntry_gen_{}.txt".format(dname, gen))
        refgen = os.path.join(ex_listdir, "getEntry_gen_{}".format(gen))
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(listdir, "{}-getEntry_prt_{}.txt".format(dname, gen))
        refprt = os.path.join(ex_listdir, "getEntry_prt_{}".format(gen))
        shutil.copyfile(refprt, outprt)
    # Create content of aldir
    ex_aldir = os.path.join(EXPPATH, "exp_aldir-pers")
    fams = [1, 4, 6, 8, 10, 11, 13, 14]
    for fam in fams:
        outgen = os.path.join(aldir, '{}-current.{}.gen').format(dname, fam)
        refgen = os.path.join(ex_aldir, "current.{}.gen").format(fam)
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(aldir, '{}-current.{}.prt').format(dname, fam)
        refprt = os.path.join(ex_aldir, "current.{}.prt").format(fam)
        shutil.copyfile(refprt, outprt)
        outmiss = os.path.join(aldir, '{}-current.{}.miss.lst').format(dname, fam)
        refmiss = os.path.join(ex_aldir, "current.{}.miss.lst").format(fam)
        shutil.copyfile(refmiss, outmiss)
        outaln = os.path.join(aldir, '{}-mafft-align.{}.aln').format(dname, fam)
        refaln = os.path.join(ex_aldir, "mafft-align.{}.aln").format(fam)
        shutil.copyfile(refaln, outaln)
        outbtr = os.path.join(aldir, '{}-mafft-prt2nuc.{}.aln').format(dname, fam)
        refbtr = os.path.join(ex_aldir, "mafft-prt2nuc.{}.aln").format(fam)
        shutil.copyfile(refbtr, outbtr)
    outmafft1 = os.path.join(aldir, '{}-mafft-align.1.aln').format(dname)
    with open(outmafft1, "w") as outf:
        outf.write("Hello !!")
    al.main(cmd, corepers, list_genomes, dname, dbpath, outdir, threads, force)
    # Check content of mafft and prt2nuc files for fam 1
    ref_mafft1 = os.path.join(ex_aldir, "mafft-align.1.aln")
    assert tutil.compare_order_content(outmafft1, ref_mafft1)
    out_p2n1 = os.path.join(aldir, '{}-mafft-prt2nuc.1.aln').format(dname)
    ref_p2n1 = os.path.join(ex_aldir, "mafft-prt2nuc.1.aln")
    assert tutil.compare_order_content(out_p2n1, ref_p2n1)
    # Check content of concatenated and grouped files
    out_concat = os.path.join(aldir, dname + "-complete.cat.aln")
    exp_concat = os.path.join(EXPPATH, "exp_pers4genome-complete.cat.aln")
    assert tutil.compare_order_content(out_concat, exp_concat)
    out_grp = os.path.join(treedir, dname + ".grp.aln")
    exp_grp = os.path.join(EXPPATH, "exp_pers4genomes.grp.aln")
    assert tutil.compare_order_content(out_grp, exp_grp)
    # Check presence of log files, and log.err is empty
    out, err = capsys.readouterr()
    logfile = os.path.join(outdir, "PanACoTA-align_TEST4-emptym1.log")
    assert os.path.isfile(logfile)
    assert os.path.isfile(logfile + ".details")
    assert os.path.isfile(logfile + ".err")
    # Check logs
    with open(logfile + ".details", "r") as lc:
        log_content = lc.readlines()
    assert ("Reading PersGenome and constructing lists of missing genomes in "
            "each family") in out
    for gen in genomes:
        assert ("For genome {0}, "
                "{2}/List-{1}/{1}-getEntry_prt_{0}.txt and {2}/List-{1}/{1}-getEntry_gen_{0}.txt "
                "already exist. The program will use them to extract proteins and genes. If you "
                "prefer to rewrite them, use option -F "
                "(or --force)".format(gen, dname, outdir)) in " ".join(log_content)
    assert ("All extraction files already existing (see detailed log for more "
            "information)") in out
    assert ("All prt and gene files for all families already exist. The program will use them "
            "for the next step. If you want to re-extract a given family, remove its prt and "
            "gen extraction files. If you want to re-extract all families, use option -F "
            "(or --force).") in " ".join(log_content)
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in out
    assert "Checking extractions for family 1" in " ".join(log_content)
    assert ("fam 1: Will redo alignment, because found a different number of proteins "
            "extracted in {0}/Align-{1}/{1}-current.1.prt (4) and proteins aligned in "
            "existing {0}/Align-{1}/"
            "{1}-mafft-align.1.aln".format(outdir, dname)) in " ".join(log_content)
    assert "Aligning family 1" in " ".join(log_content)
    assert "Back-translating family 1" in " ".join(log_content)
    fams2 = fams[1:]
    for fam in fams2:
        assert "Checking extractions for family {}".format(fam) in " ".join(log_content)
        assert ("Alignment already done for family {}. The program will use it for next "
                "steps").format(fam) in " ".join(log_content)
    assert "Concatenating all alignment files" in out
    assert "Grouping alignments per genome" in out
    assert "END" in out


def test_main_emptyprtgen(capsys):
    """
    Test main all files exist but empty prt extraction for 1 fam, empty gen extract for 1 other fam
    -> exits with error message because wrong extractions for family 1 and 4 (and not redone as
    not force). Should have removed align and prt2nuc files for fams 1 and 4, and concat and group files
    """
    corepers = os.path.join(TESTPATH, "test_pers0.99FX.lst")
    list_genomes = os.path.join("test", "data", "pangenome", "test_files", "list_to_pan.txt")
    dname = "TEST4-empty-prtgen"
    dbpath = os.path.join("test", "data", "pangenome", "test_files", "example_db")
    outdir = os.path.join(GENEPATH, "test_main_exist_empty-prtgen")
    threads = 1
    force = False
    cmd = "cmd test_main_exist_empty-m1"
    # Create output directories and files
    aldir = os.path.join(outdir, "Align-" + dname)
    listdir = os.path.join(outdir, "List-" + dname)
    treedir = os.path.join(outdir, "Phylo-" + dname)
    os.makedirs(aldir)
    os.makedirs(listdir)
    os.makedirs(treedir)
    # Create content of listdir
    ex_listdir = os.path.join(EXPPATH, "exp_listdir-pers")
    genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    for gen in genomes:
        outgen = os.path.join(listdir, "{}-getEntry_gen_{}.txt".format(dname, gen))
        refgen = os.path.join(ex_listdir, "getEntry_gen_{}".format(gen))
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(listdir, "{}-getEntry_prt_{}.txt".format(dname, gen))
        refprt = os.path.join(ex_listdir, "getEntry_prt_{}".format(gen))
        shutil.copyfile(refprt, outprt)
    # Create content of aldir
    ex_aldir = os.path.join(EXPPATH, "exp_aldir-pers")
    fams = [1, 4, 6, 8, 10, 11, 13, 14]
    for fam in fams:
        outgen = os.path.join(aldir, '{}-current.{}.gen').format(dname, fam)
        refgen = os.path.join(ex_aldir, "current.{}.gen").format(fam)
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(aldir, '{}-current.{}.prt').format(dname, fam)
        refprt = os.path.join(ex_aldir, "current.{}.prt").format(fam)
        shutil.copyfile(refprt, outprt)
        outmiss = os.path.join(aldir, '{}-current.{}.miss.lst').format(dname, fam)
        refmiss = os.path.join(ex_aldir, "current.{}.miss.lst").format(fam)
        shutil.copyfile(refmiss, outmiss)
        outaln = os.path.join(aldir, '{}-mafft-align.{}.aln').format(dname, fam)
        refaln = os.path.join(ex_aldir, "mafft-align.{}.aln").format(fam)
        shutil.copyfile(refaln, outaln)
        outbtr = os.path.join(aldir, '{}-mafft-prt2nuc.{}.aln').format(dname, fam)
        refbtr = os.path.join(ex_aldir, "mafft-prt2nuc.{}.aln").format(fam)
        shutil.copyfile(refbtr, outbtr)
    # Empty prt file for fam 1, empty gen file for fam 4
    outprt1 = os.path.join(aldir, '{}-current.1.prt').format(dname)
    outgen4 = os.path.join(aldir, '{}-current.4.gen').format(dname)
    open(outprt1, "w").close()
    open(outgen4, "w").close()
    # Empty concat and group files
    out_concat = os.path.join(aldir, dname + "-complete.cat.aln")
    open(out_concat, "w").close()
    out_grp = os.path.join(treedir, dname + ".grp.aln")
    open(out_grp, "w").close()
    # Run alignment, but problem
    with pytest.raises(SystemExit):
        al.main(cmd, corepers, list_genomes, dname, dbpath, outdir, threads, force)
    # Check that align and prt2nuc files are removed for fam 1 and 4
    for fam in [1, 4]:
        outaln = os.path.join(aldir, '{}-mafft-align.{}.aln').format(dname, fam)
        outbtr = os.path.join(aldir, '{}-mafft-prt2nuc.{}.aln').format(dname, fam)
        assert not os.path.isfile(outaln)
        assert not os.path.isfile(outbtr)
    # Check concatenated and grouped files are removed
    assert not os.path.isfile(out_concat)
    assert not os.path.isfile(out_grp)
    # Check logs
    out, err = capsys.readouterr()
    logfile = os.path.join(outdir, "PanACoTA-align_TEST4-empty-prtgen.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Reading PersGenome and constructing lists of missing genomes in "
            "each family") in out
    for gen in genomes:
        assert ("For genome {0}, "
                "{2}/List-{1}/{1}-getEntry_prt_{0}.txt and {2}/List-{1}/{1}-getEntry_gen_{0}.txt "
                "already exist. The program will use them to extract proteins and genes. If you "
                "prefer to rewrite them, use option -F "
                "(or --force)".format(gen, dname, outdir)) in " ".join(log_content)
    assert ("All extraction files already existing (see detailed log for more "
            "information)") in out
    assert ("All prt and gene files for all families already exist. The program will use them "
            "for the next step. If you want to re-extract a given family, remove its prt and "
            "gen extraction files. If you want to re-extract all families, use option -F "
            "(or --force).") in " ".join(log_content)
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in out
    assert "Checking extractions for family 1" in " ".join(log_content)
    assert ("fam 1: wrong sum of missing genomes (0) and prt extracted (0) for 4 genomes in "
            "the dataset.") in " ".join(log_content)
    assert "Checking extractions for family 4" in " ".join(log_content)
    assert ("fam 4: wrong sum of missing genomes (1) and gen extracted (0) for 4 genomes in "
            "the dataset.") in " ".join(log_content)
    fams2 = fams[2:]
    for fam in fams2:
        assert "Checking extractions for family {}".format(fam) in " ".join(log_content)
        assert ("Alignment already done for family {}. The program will use it for next "
                "steps").format(fam) in " ".join(log_content)
    assert ("At least one alignment did not run well. See detailed log file for more information. "
            "Program will stop here, alignments won't be grouped by genome.") in err


def test_main_emptylist(capsys):
    """
    Test main all files exist but wrong list of proteins for 1 genome (and no files of extracted
    prt nor gen )-> exits with error, and deleted all gen, prt, align, concat and group files
    """
    corepers = os.path.join(TESTPATH, "test_pers0.99FX.lst")
    list_genomes = os.path.join("test", "data", "pangenome", "test_files", "list_to_pan.txt")
    dname = "TEST4-wronglist"
    dbpath = os.path.join("test", "data", "pangenome", "test_files", "example_db")
    outdir = os.path.join(GENEPATH, "test_main_exist_wronglist")
    threads = 1
    force = False
    cmd = "cmd test_main_exist_empty-m1"
    # Create output directories and files
    aldir = os.path.join(outdir, "Align-" + dname)
    listdir = os.path.join(outdir, "List-" + dname)
    treedir = os.path.join(outdir, "Phylo-" + dname)
    os.makedirs(aldir)
    os.makedirs(listdir)
    os.makedirs(treedir)
    # Create content of listdir
    ex_listdir = os.path.join(EXPPATH, "exp_listdir-pers")
    genomes = ["GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    for gen in genomes:
        outgen = os.path.join(listdir, "{}-getEntry_gen_{}.txt".format(dname, gen))
        refgen = os.path.join(EXPPATH, f"exp_wronglist-getEntry_gen_{gen}.txt")
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(listdir, "{}-getEntry_prt_{}.txt".format(dname, gen))
        refprt = os.path.join(EXPPATH, f"exp_wronglist-getEntry_prt_{gen}.txt")
        shutil.copyfile(refprt, outprt)
    # for genome 1, empty lists
    outgen = os.path.join(listdir, "{}-getEntry_gen_GEN2.1017.00001.txt".format(dname))
    outprt = os.path.join(listdir, "{}-getEntry_prt_GEN2.1017.00001.txt".format(dname))
    open(outgen, "w").close()
    open(outprt, "w").close()
    # Create content of aldir
    ex_aldir = os.path.join(EXPPATH, "exp_aldir-pers")
    fams = [1, 4, 6, 8, 10, 11, 13, 14]
    for fam in fams:
        outmiss = os.path.join(aldir, '{}-current.{}.miss.lst').format(dname, fam)
        refmiss = os.path.join(ex_aldir, "current.{}.miss.lst").format(fam)
        shutil.copyfile(refmiss, outmiss)
        outaln = os.path.join(aldir, '{}-mafft-align.{}.aln').format(dname, fam)
        refaln = os.path.join(ex_aldir, "mafft-align.{}.aln").format(fam)
        shutil.copyfile(refaln, outaln)
        outbtr = os.path.join(aldir, '{}-mafft-prt2nuc.{}.aln').format(dname, fam)
        refbtr = os.path.join(ex_aldir, "mafft-prt2nuc.{}.aln").format(fam)
        shutil.copyfile(refbtr, outbtr)
    out_concat = os.path.join(aldir, dname + "-complete.cat.aln")
    # Empty concat and group files
    open(out_concat, "w").close()
    out_grp = os.path.join(treedir, dname + ".grp.aln")
    open(out_grp, "w").close()
    with pytest.raises(SystemExit):
        al.main(cmd, corepers, list_genomes, dname, dbpath, outdir, threads, force)
    # Check that prt, gen, align and prt2nuc files are removed for all fams
    for fam in fams:
        outaln = os.path.join(aldir, '{}-mafft-align.{}.aln').format(dname, fam)
        outbtr = os.path.join(aldir, '{}-mafft-prt2nuc.{}.aln').format(dname, fam)
        assert not os.path.isfile(outaln)
        assert not os.path.isfile(outbtr)
    # Check logs
    out, err = capsys.readouterr()
    logfile = os.path.join(outdir, "PanACoTA-align_TEST4-wronglist.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Reading PersGenome and constructing lists of missing genomes in "
            "each family") in out
    for gen in genomes:
        assert (f"For genome {gen}, test/data/align/generated_by_func_tests/"
                "test_main_exist_wronglist/List-TEST4-wronglist/"
                f"TEST4-wronglist-getEntry_prt_{gen}.txt and test/data/align/"
                "generated_by_func_tests/test_main_exist_wronglist/List-TEST4-wronglist/"
                f"TEST4-wronglist-getEntry_gen_{gen}.txt already exist. "
                "The program will use them to extract proteins and genes. If you prefer to "
                "rewrite them, use option -F (or --force)") in " ".join(log_content)
    assert "Extracting proteins and genes from all genomes" in out
    for gen in genomes:
        assert "Extracting proteins and genes from {}".format(gen) in " ".join(log_content)
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in out
    assert "Checking extractions for family 4" in " ".join(log_content)
    assert ("fam 4: wrong sum of missing genomes (1) and prt extracted (2) for 4 genomes in "
            "the dataset.") in " ".join(log_content)
    assert "Checking extractions for family 8" in " ".join(log_content)
    assert ("fam 8: wrong sum of missing genomes (1) and prt extracted (2) for 4 genomes in "
            "the dataset.") in " ".join(log_content)
    for fam in [1, 6, 10, 11, 13, 14]:
        assert "Checking extractions for family {}".format(fam) in " ".join(log_content)
        assert ("fam {}: wrong sum of missing genomes (0) and prt extracted (3) for 4 genomes in "
                "the dataset.".format(fam)) in " ".join(log_content)
    assert ("At least one alignment did not run well. See detailed log file for more information. "
            "Program will stop here, alignments won't be grouped by genome.") in err


def test_main_from_parse_force(capsys):
    """
    Test main all files exist but force -> redo everything
    """
    import argparse
    corepers = os.path.join(TESTPATH, "test_pers0.99FX.lst")
    list_genomes = os.path.join("test", "data", "pangenome", "test_files", "list_to_pan.txt")
    dname = "TEST4-fromparse"
    dbpath = os.path.join("test", "data", "pangenome", "test_files", "example_db")
    outdir = os.path.join(GENEPATH, "test_main_from_parse_force")
    threads = 1
    force = True
    # Create parser
    args = argparse.Namespace()
    args.corepers = corepers
    args.list_genomes = list_genomes
    args.dataset_name = dname
    args.dbpath = dbpath
    args.outdir = outdir
    args.threads = threads
    args.force = force
    args.verbose = 1
    args.quiet = False
    args.argv = "cmd test_main_exist_empty-m1"
    # Create output directories and files
    aldir = os.path.join(outdir, "Align-" + dname)
    listdir = os.path.join(outdir, "List-" + dname)
    treedir = os.path.join(outdir, "Phylo-" + dname)
    os.makedirs(aldir)
    os.makedirs(listdir)
    os.makedirs(treedir)
    # Create content of listdir
    ex_listdir = os.path.join(EXPPATH, "exp_listdir-pers")
    genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    for gen in genomes[1:]:
        outgen = os.path.join(listdir, f"{dname}-getEntry_gen_{gen}.txt")
        refgen = os.path.join(EXPPATH, f"exp_wronglist-getEntry_gen_{gen}.txt")
        shutil.copyfile(refgen, outgen)
        outprt = os.path.join(listdir, "{}-getEntry_prt_{}.txt".format(dname, gen))
        refprt = os.path.join(EXPPATH, f"exp_wronglist-getEntry_prt_{gen}.txt")
        shutil.copyfile(refprt, outprt)
    # for genome 1, empty lists
    outgen = os.path.join(listdir, "{}-getEntry_gen_GEN2.1017.00001.txt".format(dname))
    outprt = os.path.join(listdir, "{}-getEntry_prt_GEN2.1017.00001.txt".format(dname))
    open(outgen, "w").close()
    open(outprt, "w").close()
    # Create content of aldir
    ex_aldir = os.path.join(EXPPATH, "exp_aldir-pers")
    fams = [1, 4, 6, 8, 10, 11, 13, 14]
    for fam in fams:
        outmiss = os.path.join(aldir, '{}-current.{}.miss.lst').format(dname, fam)
        refmiss = os.path.join(ex_aldir, "current.{}.miss.lst").format(fam)
        shutil.copyfile(refmiss, outmiss)
        outaln = os.path.join(aldir, '{}-mafft-align.{}.aln').format(dname, fam)
        refaln = os.path.join(ex_aldir, "mafft-align.{}.aln").format(fam)
        shutil.copyfile(refaln, outaln)
        outbtr = os.path.join(aldir, '{}-mafft-prt2nuc.{}.aln').format(dname, fam)
        refbtr = os.path.join(ex_aldir, "mafft-prt2nuc.{}.aln").format(fam)
        shutil.copyfile(refbtr, outbtr)
    out_concat = os.path.join(aldir, dname + "-complete.cat.aln")
    # Empty concat and group files
    open(out_concat, "w").close()
    out_grp = os.path.join(treedir, dname + ".grp.aln")
    open(out_grp, "w").close()
    # Run align
    al.main_from_parse(args)
    # Check content of listdir
    genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    for gen in genomes:
        assert os.path.isfile(os.path.join(listdir, f"{dname}-getEntry_gen_{gen}.txt"))
        assert os.path.isfile(os.path.join(listdir, f"{dname}-getEntry_prt_{gen}.txt"))
    # Check content of aldir
    fams = [1, 4, 6, 8, 10, 11, 13, 14]
    for fam in fams:
        assert os.path.isfile(os.path.join(aldir, '{}-current.{}.gen').format(dname, fam))
        assert os.path.isfile(os.path.join(aldir, '{}-current.{}.prt').format(dname, fam))
        assert os.path.isfile(os.path.join(aldir, '{}-current.{}.miss.lst').format(dname, fam))
        assert os.path.isfile(os.path.join(aldir, '{}-mafft-align.{}.aln').format(dname, fam))
        assert os.path.isfile(os.path.join(aldir, '{}-mafft-prt2nuc.{}.aln').format(dname, fam))
    out_concat = os.path.join(aldir, dname + "-complete.cat.aln")
    exp_concat = os.path.join(EXPPATH, "exp_pers4genome-complete.cat.aln")
    assert tutil.compare_order_content(out_concat, exp_concat)
    # Check content of treedir
    out_grp = os.path.join(treedir, dname + ".grp.aln")
    exp_grp = os.path.join(EXPPATH, "exp_pers4genomes.grp.aln")
    assert tutil.compare_order_content(out_grp, exp_grp)
    # Check logs
    out, err = capsys.readouterr()
    logfile = os.path.join(outdir, "PanACoTA-align_TEST4-fromparse.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Reading PersGenome and constructing lists of missing genomes in "
            "each family") in out
    assert "Extracting proteins and genes from all genomes" in out
    for gen in genomes:
        assert "Extracting proteins and genes from {}".format(gen) in " ".join(log_content)
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in out
    for fam in fams:
        assert "Checking extractions for family {}".format(fam) in " ".join(log_content)
        assert "Aligning family {}".format(fam) in " ".join(log_content)
        assert "Back-translating family {}".format(fam) in " ".join(log_content)
    assert "Concatenating all alignment files" in out
    assert "Grouping alignments per genome" in out
    assert "END" in out



#!/usr/bin/env python3
# coding: utf-8

"""
Functional tests for the parser of 'corepers' subcommand
"""

import pytest
import argparse

from PanACoTA.subcommands import corepers


def test_parser_noarg(capsys):
    """
    Test that when the script is called without any argument, an error message appears,
    indicating the required arguments.
    """
    parser = argparse.ArgumentParser(description="Do corepers", add_help=False)
    corepers.build_parser(parser)
    with pytest.raises(SystemExit):
        corepers.parse(parser, "".split())
    _, err = capsys.readouterr()
    assert "usage: " in err
    assert " -p PANGENOME -o OUTPUTDIR [-t TOL] [-M] [-X] [-F]" in err
    assert "[-v] [-q]" in err
    assert "[-h]" in err
    assert "the following arguments are required: -p, -o" in err


def test_parser_tol_nofloat(capsys):
    """
    Test that when the '-t tol' parameter given is not a number, it raises
    the expected error message
    """
    parser = argparse.ArgumentParser(description="Do corepers", add_help=False)
    corepers.build_parser(parser)
    with pytest.raises(SystemExit):
        corepers.parse(parser, "-p pangenome -t mytol".split())
    _, err = capsys.readouterr()
    assert "argument -t tol: invalid float value: mytol" in err


def test_parser_tol_neg(capsys):
    """
    Test that when the '-t tol' parameter given is a negative number, it raises
    the expected error message
    """
    parser = argparse.ArgumentParser(description="Do corepers", add_help=False)
    corepers.build_parser(parser)
    with pytest.raises(SystemExit):
        corepers.parse(parser, "-p pangenome -t -0.5".split())
    _, err = capsys.readouterr()
    assert ("The minimum %% of genomes required in a family to be persistent must be in [0, "
            "1]. Invalid value: -0.5") in err


def test_parser_multi_mixed(capsys):
    """
    Test that when the user asks for multi (-M) and mixed (-X) persistent
    genome, it returns an error message
    """
    parser = argparse.ArgumentParser(description="Do corepers", add_help=False)
    corepers.build_parser(parser)
    with pytest.raises(SystemExit):
        corepers.parse(parser, "-p pangenome -o toto -M -X".split())
    _, err = capsys.readouterr()
    assert ("-M and -X options cannot be activated together. Choose if you want to:\n"
            "- allow several members in any number of genomes of a family (-M)\n"
            "- allow several members in only '1-tol'% of the genomes of a family "
            "(other 'tol'% genomes must have exactly 1 member) (-X)") in err


def test_parser_mixed_tol1(capsys):
    """
    Test that when the users asks for a mixed persistent genome with 100% of
    strains with exactly 1 member, it returns an error message
    """
    parser = argparse.ArgumentParser(description="Do corepers", add_help=False)
    corepers.build_parser(parser)
    with pytest.raises(SystemExit):
        corepers.parse(parser, "-p pangenome -t 1 -X -o toto".split())
    _, err = capsys.readouterr()
    assert ("You are asking for mixed families, while asking for 100% of the genomes of "
            "a family to have exactly one member, which is not compatible. Do you want "
            "to \n- lower the percentage of genomes required to have exactly "
            "1 member (-t tol)\n- not allow mixed families (remove -X option)") in err


def test_parser_floor_tol1(capsys):
    """
    Test that when the user asks for 100% of strains having exactly 1 member,
    + the option to use 'floor', it returns an error message.
    """
    parser = argparse.ArgumentParser(description="Do corepers", add_help=False)
    corepers.build_parser(parser)
    with pytest.raises(SystemExit):
        corepers.parse(parser, "-p pangenome -t 1 -F -o outdir".split())
    _, err = capsys.readouterr()
    assert ("You are asking to use floor('tol'*N) as a minimum number of genomes "
            "present in a family, but with 'tol'=1: the minimum number of genomes "
            "will always be equal to N, using floor or the default ceil! Either "
            "use a 'tol' lower than 1, or remove the '-F' option.") in err


def test_parser_default():
    """
    Test that when run with the minimum required arguments, all default values are
    as expected.
    """
    parser = argparse.ArgumentParser(description="Do corepers", add_help=False)
    corepers.build_parser(parser)
    options = corepers.parse(parser, "-p pangenome -o outdir".split())
    assert options.pangenome == "pangenome"
    assert options.tol == 1
    assert options.multi is False
    assert options.mixed is False
    assert options.outputdir == "outdir"
    assert not options.floor
    assert options.verbose == 0
    assert not options.quiet


def test_parser_mixed_floor():
    """
    Test that when run with tol of 0.99, mixed, and use floor, it returns
    the expected values for all arguments
    """
    parser = argparse.ArgumentParser(description="Do corepers", add_help=False)
    corepers.build_parser(parser)
    options = corepers.parse(parser, "-p pangenome --tol 0.99 -X -F -o outdir".split())
    assert options.pangenome == "pangenome"
    assert options.tol == 0.99
    assert options.multi is False
    assert options.mixed is True
    assert options.outputdir == "outdir"
    assert options.floor is True
    assert options.verbose == 0
    assert not options.quiet



#!/usr/bin/env python3

"""
Functional tests for pangenome subcommand
"""
import glob
import os
import shutil
import subprocess
import pytest

import PanACoTA.subcommands.pangenome as pan

LOGFILE_BASE = "func_test_pangenome"
PAN_DIR = os.path.join("test", "data", "pangenome")
TEST_FILES = os.path.join(PAN_DIR, "test_files")
DBPATH = os.path.join(TEST_FILES, "example_db", "Proteins")
EXP_FILES = os.path.join(PAN_DIR, "exp_files")
GENEPATH = os.path.join(PAN_DIR, "generated_by_func-tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module
    """
    # Init logger to level detail (15)
    # utils.init_logger(LOGFILE_BASE, 0, 'test_func_pan', verbose=1)
    if not os.path.isdir(GENEPATH):
        os.mkdir(GENEPATH)
    print("setup")

    yield
    # os.remove(LOGFILE_BASE + ".log")
    # os.remove(LOGFILE_BASE + ".log.details")
    # os.remove(LOGFILE_BASE + ".log.err")
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("Removed log files")


def test_main_from_parse():
    """
    Test main when we give the output of the parser
    """
    import argparse
    lstinfo = os.path.join(TEST_FILES, "list_to_pan.txt")
    name = "testFromParsePAN4"
    min_id = 0.8
    args = argparse.Namespace()
    args.lstinfo_file = lstinfo
    args.dataset_name = name
    args.dbpath = os.path.join(GENEPATH, "database")
    # copy db_path folder to output folder, as it will modify it
    shutil.copytree(DBPATH, args.dbpath)
    args.min_id = min_id
    args.outdir = GENEPATH
    args.clust_mode = 1
    args.spedir = None
    args.threads = 1
    args.outfile = None
    args.verbose = 0
    args.quiet = False
    args.argv = ["pangenome", "pan.py", "test_main_from_parse"]
    # Run main_from_parse
    pan.main_from_parse(args)
    # Check rprt bank was created, and in expected location
    prtbank = os.path.join(args.dbpath, name + ".All.prt")
    assert os.path.isfile(prtbank)

    # Check presence of tmp folder
    tmp_base = os.path.join(GENEPATH, "tmp_testFromParsePAN4.All.prt_0.8-mode1_*")
    assert len(glob.glob(tmp_base)) == 1
    # check presence of mmseq cluster files
    cluster = os.path.join(GENEPATH, name + ".All.prt-clust-0.8-mode1_*")
    clust_files = glob.glob(cluster)
    assert len(clust_files) == 4
    # Check presence of pangenome files (pangenome, matrices, summary)
    pan_files = glob.glob(os.path.join(GENEPATH, "PanGenome-testFromParsePAN4*"))
    to_check = [".tsv.lst", ".tsv.lst.quali.txt", ".tsv.lst.quanti.txt", ".tsv.lst.summary.txt"]
    found = []
    pangenome_file = ""
    for f in pan_files:
        for c in to_check:
            if f.endswith(c):
                found.append(c)
            if f.endswith(".tsv.lst"):
                pangenome_file = f
    assert set(found) == set(to_check)
    # Check content of pangenome
    exp_pan = os.path.join(EXP_FILES, "exp_pangenome-4genomes.lst")
    # Check that all families are as expected. Compare lines without the family number
    with open(exp_pan, "r") as ep, open(pangenome_file, "r") as panf:
        lines_exp = []
        lines_out = []
        for line_exp in ep:
            lines_exp.append(" ".join(line_exp.split()[1:]))
        for line_out in panf:
            lines_out.append(" ".join(line_out.split()[1:]))
    assert len(lines_exp) == len(lines_out)
    assert set(lines_exp) == set(lines_out)


def test_main(caplog):
    """
    Test that from empty directory, it creates all expected files and
    returns correct logs
    """
    lstinfo = os.path.join(TEST_FILES, "list_to_pan.txt")
    name = "testPAN4"
    min_id = 0.8
    outdir = GENEPATH
    clust_mode = 1
    spe_dir = None
    threads = 1
    cmd = "cmd"
    used_dbpath = os.path.join(GENEPATH, "database")
    # copy db_path folder to output folder, as it will modify it
    shutil.copytree(DBPATH, used_dbpath)
    out_panfile = os.path.join(outdir, "PanGenome-testPAN4.All.prt-clust-0.8-mode1_")
    assert pan.main(cmd, lstinfo, name, used_dbpath, min_id, outdir, clust_mode,
                    spe_dir, threads, verbose=2).startswith(out_panfile)
    # Checl creation of prt bank
    prtbank = os.path.join(used_dbpath, "testPAN4.All.prt")
    assert os.path.isfile(prtbank)
    # Check presence of tmp folder
    tmp_base = os.path.join(outdir, "tmp_testPAN4.All.prt_0.8-mode1_*")
    assert len(glob.glob(tmp_base)) == 1
    # Check presence of pangenome files (pangenome, matrices, summary)
    pan_files = glob.glob(os.path.join(GENEPATH, "PanGenome-testPAN4*"))
    to_check = [".tsv.lst", ".tsv.lst.quali.txt", ".tsv.lst.quanti.txt", ".tsv.lst.summary.txt"]
    found = []
    pangenome_file = ""
    for f in pan_files:
        for c in to_check:
            if f.endswith(c):
                found.append(c)
            if f.endswith(".tsv.lst"):
                pangenome_file = f
    assert set(found) == set(to_check)
    # Check content of pangenome
    exp_pan = os.path.join(EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(pangenome_file, "r") as panf:
        lines_exp = []
        lines_out = []
        for line_exp in ep:
            lines_exp.append(" ".join(line_exp.split()[1:]))
        for line_out in panf:
            lines_out.append(" ".join(line_out.split()[1:]))
    assert len(lines_exp) == len(lines_out)
    assert set(lines_exp) == set(lines_out)
    # Check log content
    assert ("Building bank with all proteins to test/data/pangenome/"
            "generated_by_func-tests/database/testPAN4.All.prt") in caplog.text
    assert "Creating database" in caplog.text
    assert "Clustering proteins..." in caplog.text
    assert "Converting mmseqs results to pangenome file" in caplog.text
    assert "Pangenome has 16 families" in caplog.text
    assert "Retrieving information from pan families" in caplog.text
    assert "Saving all information to a binary file for later use" in caplog.text
    assert "Generating qualitative and quantitative matrix, and summary file" in caplog.text


def test_main_prt_exist(caplog):
    """
    Test that when the prt bank already exists, it writes it to the logs, and
    continues using it.
    """
    lstinfo = os.path.join(TEST_FILES, "list_to_pan.txt")
    name = "test2PAN4"
    min_id = 0.8
    outdir = GENEPATH
    clust_mode = 1
    spe_dir = None
    threads = 1
    cmd = "cmd"
    used_dbpath = os.path.join(GENEPATH, "database")
    # copy db_path folder to output folder, as it will modify it
    shutil.copytree(DBPATH, used_dbpath)

    # Copy prt bank to database folder, so that it is not created again
    src_prt_bank = os.path.join(EXP_FILES, "exp_EXEM.All.prt")
    dest_prt_bank = os.path.join(used_dbpath, "test2PAN4.All.prt")
    shutil.copyfile(src_prt_bank, dest_prt_bank)

    out_panfile = os.path.join(outdir, "PanGenome-test2PAN4.All.prt-clust-0.8-mode1_")
    assert pan.main(cmd, lstinfo, name, used_dbpath, min_id, outdir, clust_mode, spe_dir,
                    threads, verbose=15).startswith(out_panfile)

    # Check presence of mmseq DB files
    msdb = os.path.join(GENEPATH, "test2PAN4.All.prt-msDB")
    assert os.path.isfile(msdb)
    assert os.path.isfile(msdb + ".index")
    assert os.path.isfile(msdb + ".lookup")
    assert os.path.isfile(msdb + "_h")
    assert os.path.isfile(msdb + "_h.index")
    # Check presence of mmseq cluster files
    cluster = os.path.join(outdir, "test2PAN4.All.prt-clust-0.8-mode1_*")
    clust_files = glob.glob(cluster)
    assert len(clust_files) == 4
    # Check presence of pangenome files (pangenome, matrices, summary)
    pan_files = glob.glob(os.path.join(GENEPATH, "PanGenome-test2PAN4*"))
    to_check = [".tsv.lst", ".tsv.lst.quali.txt", ".tsv.lst.quanti.txt", ".tsv.lst.summary.txt"]
    found = []
    pangenome_file = ""
    for f in pan_files:
        for c in to_check:
            if f.endswith(c):
                found.append(c)
            if f.endswith(".tsv.lst"):
                pangenome_file = f
    assert set(found) == set(to_check)
    # Check presence of tmp folder
    tmp_base = os.path.join(outdir, "tmp_test2PAN4.All.prt_0.8-mode1_*")
    assert len(glob.glob(tmp_base)) == 1
    # Check content of pangenome
    exp_pan = os.path.join(EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(pangenome_file, "r") as panf:
        lines_exp = []
        lines_out = []
        for line_exp in ep:
            lines_exp.append(" ".join(line_exp.split()[1:]))
        for line_out in panf:
            lines_out.append(" ".join(line_out.split()[1:]))
    assert len(lines_exp) == len(lines_out)
    assert set(lines_exp) == set(lines_out)
    # Check log content
    assert ("Protein bank test/data/pangenome/generated_by_func-tests/"
            "database/test2PAN4.All.prt already exists. It will "
            "be used by mmseqs.") in caplog.text
    assert "Creating database" in caplog.text
    assert "Clustering proteins..." in caplog.text
    assert "Converting mmseqs results to pangenome file" in caplog.text
    assert "Pangenome has 16 families" in caplog.text
    assert "Retrieving information from pan families" in caplog.text
    assert "Saving all information to a binary file for later use" in caplog.text
    assert "Generating qualitative and quantitative matrix, and summary file" in caplog.text


def test_main_spedir(caplog):
    """
    Test that from empty directory, it creates all expected files and
    returns correct logs
    """
    lstinfo = os.path.join(TEST_FILES, "list_to_pan.txt")
    name = "test3PAN4"
    min_id = 0.8
    outdir = GENEPATH
    clust_mode = 1
    spe_dir = os.path.join(GENEPATH, "spedir")
    threads = 1
    cmd = "cmd"
    used_dbpath = os.path.join(GENEPATH, "database")
    # copy db_path folder to output folder, as it will modify it
    shutil.copytree(DBPATH, used_dbpath)

    out_panfile = os.path.join(outdir, "PanGenome-test3PAN4.All.prt-clust-0.8-mode1_")
    assert pan.main(cmd, lstinfo, name, used_dbpath, min_id, outdir, clust_mode, spe_dir,
                    threads, verbose=15).startswith(out_panfile)
    # Checl creation of prt bank
    prtbank = os.path.join(spe_dir, "test3PAN4.All.prt")
    assert os.path.isfile(prtbank)
    # Check presence of mmseq DB files
    msdb = os.path.join(outdir, "test3PAN4.All.prt-msDB")
    assert os.path.isfile(msdb)
    assert os.path.isfile(msdb + ".index")
    assert os.path.isfile(msdb + ".lookup")
    assert os.path.isfile(msdb + "_h")
    assert os.path.isfile(msdb + "_h.index")
    # Check presence of mmseq cluster files
    cluster = os.path.join(outdir, "test3PAN4.All.prt-clust-0.8-mode1_*")
    clust_files = glob.glob(cluster)
    assert len(clust_files) == 4
    # Check presence of tmp folder
    tmp_base = os.path.join(outdir, "tmp_test3PAN4.All.prt_0.8-mode1_*")
    assert len(glob.glob(tmp_base)) == 1
    # Check presence of pangenome files (pangenome, matrices, summary)
    pan_files = glob.glob(os.path.join(GENEPATH, "PanGenome-test3PAN4*"))
    to_check = [".tsv.lst", ".tsv.lst.quali.txt", ".tsv.lst.quanti.txt", ".tsv.lst.summary.txt"]
    found = []
    pangenome_file = ""
    for f in pan_files:
        for c in to_check:
            if f.endswith(c):
                found.append(c)
            if f.endswith(".tsv.lst"):
                pangenome_file = f
    assert set(found) == set(to_check)
    # Check content of pangenome
    exp_pan = os.path.join(EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(pangenome_file, "r") as panf:
        lines_exp = []
        lines_out = []
        for line_exp in ep:
            lines_exp.append(" ".join(line_exp.split()[1:]))
        for line_out in panf:
            lines_out.append(" ".join(line_out.split()[1:]))
    assert len(lines_exp) == len(lines_out)
    assert set(lines_exp) == set(lines_out)
    # Check log content
    assert ("Building bank with all proteins to test/data/pangenome/"
            "generated_by_func-tests/spedir/test3PAN4.All.prt") in caplog.text
    assert "Creating database" in caplog.text
    assert "Clustering proteins..." in caplog.text
    assert "Converting mmseqs results to pangenome file" in caplog.text
    assert "Pangenome has 16 families" in caplog.text
    assert "Retrieving information from pan families" in caplog.text
    assert "Saving all information to a binary file for later use" in caplog.text
    assert "Generating qualitative and quantitative matrix, and summary file" in caplog.text


def test_main_outfile(caplog):
    """
    Test that when giving a name for pangenome file, it creates expected files
    """
    lstinfo = os.path.join(TEST_FILES, "list_to_pan.txt")
    name = "test4PAN4"
    min_id = 0.8
    outdir = GENEPATH
    clust_mode = 1
    spe_dir = None
    threads = 1
    cmd = "cmd"
    outfile = "my_pangenome"
    used_dbpath = os.path.join(GENEPATH, "database")
    # copy db_path folder to output folder, as it will modify it
    shutil.copytree(DBPATH, used_dbpath)

    assert pan.main(cmd, lstinfo, name, used_dbpath, min_id, outdir, clust_mode, spe_dir,
                    threads, outfile=outfile) == os.path.join(outdir, outfile)

    prtbank = os.path.join(used_dbpath, "test4PAN4.All.prt")
    assert os.path.isfile(prtbank)
    # Check presence of mmseq DB files
    msdb = os.path.join(outdir, "test4PAN4.All.prt-msDB")
    assert os.path.isfile(msdb)
    assert os.path.isfile(msdb + ".index")
    assert os.path.isfile(msdb + ".lookup")
    assert os.path.isfile(msdb + "_h")
    assert os.path.isfile(msdb + "_h.index")
    # Check presence of mmseq cluster files
    cluster = os.path.join(outdir, "test4PAN4.All.prt-clust-0.8-mode1_*")
    clust_files = glob.glob(cluster)
    assert len(clust_files) == 4
    # Check presence of tmp folder
    tmp_base = os.path.join(outdir, "tmp_test4PAN4.All.prt_0.8-mode1_*")
    assert len(glob.glob(tmp_base)) == 1
    # Check presence of pangenome files (pangenome, matrices, summary)
    outf = os.path.join(outdir, outfile)
    assert os.path.isfile(outf)
    assert os.path.isfile(outf + ".quali.txt")
    assert os.path.isfile(outf + ".quanti.txt")
    assert os.path.isfile(outf + ".summary.txt")
    # Check content of pangenome
    exp_pan = os.path.join(EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(outf, "r") as panf:
        lines_exp = []
        lines_out = []
        for line_exp in ep:
            lines_exp.append(" ".join(line_exp.split()[1:]))
        for line_out in panf:
            lines_out.append(" ".join(line_out.split()[1:]))
    assert len(lines_exp) == len(lines_out)
    assert set(lines_exp) == set(lines_out)
    # Check log content
    assert ("Building bank with all proteins to test/data/pangenome/"
            "generated_by_func-tests/database/test4PAN4.All.prt") in caplog.text
    assert "Creating database" in caplog.text
    assert "Clustering proteins..." in caplog.text
    assert "Converting mmseqs results to pangenome file" in caplog.text
    assert "Pangenome has 16 families" in caplog.text
    assert "Retrieving information from pan families" in caplog.text
    assert "Saving all information to a binary file for later use" in caplog.text
    assert "Generating qualitative and quantitative matrix, and summary file" in caplog.text


def test_pangenome_all():
    """
    Test when calling pangenome from command line, it runs and gives expected output files
    """
    lstinfo = os.path.join(TEST_FILES, "list_to_pan.txt")
    name = "testAllPAN4"
    min_id = 0.8
    outdir = GENEPATH
    clust_mode = 1
    spe_dir = None
    threads = 1
    used_dbpath = os.path.join(GENEPATH, "database")
    # copy db_path folder to output folder, as it will modify it
    shutil.copytree(DBPATH, used_dbpath)

    cmd = f"PanACoTA pangenome -l {lstinfo} -n {name} -d {used_dbpath} -o {outdir} -vv"
    ret = subprocess.call(cmd.split())
    assert ret == 0

    prtbank = os.path.join(used_dbpath, "testAllPAN4.All.prt")
    assert os.path.isfile(prtbank)
    # Check presence of mmseq DB files
    msdb = os.path.join(outdir, "testAllPAN4.All.prt-msDB")
    assert os.path.isfile(msdb)
    assert os.path.isfile(msdb + ".index")
    assert os.path.isfile(msdb + ".lookup")
    assert os.path.isfile(msdb + "_h")
    assert os.path.isfile(msdb + "_h.index")
    # Check presence of mmseq cluster files
    cluster = os.path.join(outdir, "testAllPAN4.All.prt-clust-0.8-mode1_*")
    clust_files = glob.glob(cluster)
    assert len(clust_files) == 4
    # Check presence of tmp folder
    tmp_base = os.path.join(outdir, "tmp_testAllPAN4.All.prt_0.8-mode1_*")
    assert len(glob.glob(tmp_base)) == 1
    # Check presence of pangenome files (pangenome, matrices, summary)
    pan_files = glob.glob(os.path.join(GENEPATH, "PanGenome-testAllPAN4*"))
    to_check = [".tsv.lst", ".tsv.lst.quali.txt", ".tsv.lst.quanti.txt", ".tsv.lst.summary.txt"]
    found = []
    pangenome_file = ""
    for f in pan_files:
        for c in to_check:
            if f.endswith(c):
                found.append(c)
            if f.endswith(".tsv.lst"):
                pangenome_file = f
    assert set(found) == set(to_check)
    # Check content of pangenome
    exp_pan = os.path.join(EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(pangenome_file, "r") as panf:
        lines_exp = []
        lines_out = []
        for line_exp in ep:
            lines_exp.append(" ".join(line_exp.split()[1:]))
        for line_out in panf:
            lines_out.append(" ".join(line_out.split()[1:]))
    assert len(lines_exp) == len(lines_out)
    assert set(lines_exp) == set(lines_out)
    # Check presence of log files, and that .err is empty
    log_base = os.path.join(outdir, "PanACoTA-pangenome_testAllPAN4.log")
    assert os.path.isfile(log_base)
    assert os.path.isfile(log_base + ".details")
    assert os.path.isfile(log_base + ".err")
    with open(log_base + ".err") as errf:
        lines = errf.readlines()
    assert lines == []



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for utils.py
"""

import pytest
import os
import logging
import shutil

import test.test_unit.utilities_for_tests as tutil
import PanACoTA.annotate_module.genome_seq_functions as gfunc

import matplotlib
matplotlib.use('AGG')

# Define variables used by several tests
DBDIR = os.path.join("test", "data", "annotate")
GEN_PATH = os.path.join(DBDIR, "genomes")
TMP_PATH = os.path.join(DBDIR, "tmp_files")
EXP_DIR = os.path.join(DBDIR, 'exp_files')
BASELINE_DIR = os.path.abspath(os.path.join(EXP_DIR, "baseline"))
GENEPATH = os.path.join(DBDIR, "generated_by_unit-tests")
logger = logging.getLogger('test_genome_func')


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    # utils.init_logger(LOGFILE_BASE, 0, 'test_postalign', verbose=1)
    if os.path.isdir(GENEPATH):
        content = os.listdir(GENEPATH)
        for f in content:
            assert f.startswith(".fuse")
    else:
        os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")



def test_calc_l90_exact():
    """
    Calculate L90 according to the given genome size and contig sizes
    2 contigs get exactly 90% of the genome
    """
    cont_size = {1: 3, 2: 800, 3: 100, 4: 90, 5: 7}
    l90 = gfunc.calc_l90(cont_size)
    assert l90 == 2


def test_calc_l90_more():
    """
    Calculate L90 according to the given genome size and contig sizes
    3 contigs get exactly more than 90%, but 2 contigs get less -> l90 = 3
    """
    cont_size = {1: 3, 2: 800, 3: 90, 4: 90, 5: 17}
    l90 = gfunc.calc_l90(cont_size)
    assert l90 == 3


def test_calc_l90_error():
    """
    Calculate L90 according to the given genome size and contig sizes
    3 contigs get exactly more than 90%, but 2 contigs get less -> l90 = 3
    """
    cont_size = {}
    assert not gfunc.calc_l90(cont_size)


def test_rename_genomes():
    """
    From a list of genomes ({genome: [name.date, path, path_to_seq, gsize, nbcont, L90]}),
    order them by species, and by decreasing quality (L90, nb_cont), and rename them,
    as well as their contigs.
    """
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta", "genome4.fasta",
          "genome5.fasta", "genome6.fasta", "genome7.fasta"]

    genomes = {gs[0]: ["SAEN.1113", os.path.join(GEN_PATH, gs[0]), "pathtoseq1", 51, 4, 2],
               gs[1]: ["SAEN.1114", os.path.join(GEN_PATH, gs[1]), "pathToSeq2", 67, 3, 3],
               gs[2]: ["ESCO.0416", os.path.join(GEN_PATH, gs[2]), "pathToSeq3", 70, 4, 1],
               gs[3]: ["ESCO.0216", os.path.join(GEN_PATH, gs[3]), "pathToSeq4", 114, 5, 2],
               gs[4]: ["SAEN.1115", os.path.join(GEN_PATH, gs[4]), "path_to_seq5", 106, 3, 1],
               gs[5]: ["ESCO.0216", os.path.join(GEN_PATH, gs[5]), "pathtoseq6", 116, 4, 2],
               gs[6]: ["SAEN.1115", os.path.join(GEN_PATH, gs[6]), "pathtoseq7", 137, 3, 2]}
    gfunc.rename_all_genomes(genomes)
    # SAEN genomes 1 and 2 have same characteristics. Their place will be chosen randomly,
    # so take into account both choices
    exp_genomes = {gs[0]: ["SAEN.1113.00003",
                           os.path.join(GEN_PATH, gs[0]), "pathtoseq1", 51, 4, 2],
                   gs[1]: ["SAEN.1114.00004",
                           os.path.join(GEN_PATH, gs[1]), "pathToSeq2", 67, 3, 3],
                   gs[2]: ["ESCO.0416.00001",
                           os.path.join(GEN_PATH, gs[2]), "pathToSeq3", 70, 4, 1],
                   gs[3]: ["ESCO.0216.00003",
                           os.path.join(GEN_PATH, gs[3]), "pathToSeq4", 114, 5, 2],
                   gs[4]: ["SAEN.1115.00001",
                           os.path.join(GEN_PATH, gs[4]), "path_to_seq5", 106, 3, 1],
                   gs[5]: ["ESCO.0216.00002",
                           os.path.join(GEN_PATH, gs[5]), "pathtoseq6", 116, 4, 2],
                   gs[6]: ["SAEN.1115.00002",
                           os.path.join(GEN_PATH, gs[6]), "pathtoseq7", 137, 3, 2]}
    assert genomes == exp_genomes


def test_split_contig_nocut():
    """
    Test that when a contig must not be cut, it returns the current number of contigs + 1
    (we just add this contig, which is not split), and writes
    this new contig to the new file
    """
    pat = None
    whole_seq = "AACTGCTTTTTAAGCGCGCTCCTGCGNNNNNGGTTGTGTGGGCCCAGAGCGAGNCG"
    cur_contig_name = ">my_contig name for my_sequence"
    contig_sizes = {"contig_1": 10}
    resfile = os.path.join(GENEPATH, "test_split_contig_nocut.fna")
    gresf = open(resfile, "w")
    num = 2

    num = gfunc.split_contig(pat, whole_seq, cur_contig_name, contig_sizes, gresf, num)
    gresf.close()

    exp_file = os.path.join(EXP_DIR, "exp_split_contig_nocut.fna")
    assert num == 3
    assert os.path.exists(resfile)
    assert tutil.compare_order_content(resfile, exp_file)


def test_split_contig_cut():
    """
    Test that when a contig must not be cut at each stretch of 3 'N', and the sequence
    contains 1 stretch of 5 'N', it cuts the sequence into 2 contigs, ignoring all 'N' of the
    stretch (even if there are more than 3 'N'). Then, it returns the current number of contigs + 2
    and writes those new contigs to the new file
    """
    pat = "NNN+"
    whole_seq = "AACTGCTTTTTAAGCGCGCTCCTGCGNNNNNGGTTGTGTGGGCCCAGAGCGAGNCG"
    cur_contig_name = ">my_contig_name for_my_sequence"
    contig_sizes = {">contig_1": 10}
    resfile = os.path.join(GENEPATH, "test_split_contig_nocut.fna")
    gresf = open(resfile, "w")
    num = 2

    num = gfunc.split_contig(pat, whole_seq, cur_contig_name, contig_sizes, gresf, num)
    gresf.close()

    exp_file = os.path.join(EXP_DIR, "exp_split_contig_cut3N.fna")
    assert num == 4
    assert os.path.exists(resfile)
    assert tutil.compare_order_content(resfile, exp_file)


def test_split_empty_contig():
    """
    Test that when we want to split an empty contig (for exemple, if the sequence starts with
    'NNNN', it returns 2 contigs: 1 empty and 1 with the rest of the sequence), this contig is
    just ignored (not written nor taken into account in contig num).
    """
    pat = "NNN+"
    whole_seq = "NNNNNAACTGCTTTTTAAGCGCGCTCCTGCGNGGTTGTGTGGGCCCAGAGCGAGNCG"
    cur_contig_name = ">my_contig_name_for_my_sequence"
    contig_sizes = {"contig_1": 10}
    resfile = os.path.join(GENEPATH, "test_split_contig_nocut.fna")
    gresf = open(resfile, "w")
    num = 2

    num = gfunc.split_contig(pat, whole_seq, cur_contig_name, contig_sizes, gresf, num)
    gresf.close()

    exp_file = os.path.join(EXP_DIR, "exp_split_empty_contig.fna")
    assert num == 3
    assert os.path.exists(resfile)
    assert tutil.compare_order_content(resfile, exp_file)


def test_format_contig_cut():
    """
    For a given contig, if we want to annotate it, and cut at each stretch of 5 'N'
    check that it writes this contig, split, in the expected file
    """
    cut = True
    pat = 'NNNNN+'
    cur_seq = "AACTGCTTTTTAAGCGCGCTCCTGCGNNNNNGGTTGTGTGGGCCCAGAGCGAGNCG"
    cur_contig_name = ">my_contig_name for_my_sequence"
    contig_sizes = {}
    resfile = os.path.join(GENEPATH, "test_format_cont_cut5N.fna")
    gresf = open(resfile, "w")
    num = 2

    assert gfunc.format_contig(cut, pat, cur_seq, cur_contig_name, "genome", contig_sizes, gresf,
                               num, logger=None) == 4
    gresf.close()

    exp_file = os.path.join(EXP_DIR, "exp_split_contig_cut3N.fna")
    assert os.path.exists(resfile)
    assert tutil.compare_order_content(resfile, exp_file)
    assert contig_sizes == {">2_my_contig_name for_my_sequence\n": 26,
                            ">3_my_contig_name for_my_sequence\n": 25}


def test_format_contig_nocut():
    """
    For a given contig, if we want to annotate it with prokka, and do not cut at each stretch of
    5 'N'check that it writes this contig as given
    """
    cut = False
    pat = None
    cur_seq = "AACTGCTTTTTAAGCGCGCTCCTGCGNNNNNGGTTGTGTGGGCCCAGAGCGAGNCG"
    cur_contig_name = ">my_contig_name_for_my_sequence\n"
    contig_sizes = {}
    resfile = os.path.join(GENEPATH, "test_format_cont_nocut_prokka.fna")
    gresf = None
    num = 2

    assert gfunc.format_contig(cut, pat, cur_seq, cur_contig_name, "genome", contig_sizes, gresf,
                               num, logger=None) == 2

    exp_file = os.path.join(EXP_DIR, "exp_split_contig_nocut.fna")
    assert not os.path.exists(resfile)
    assert contig_sizes == {">my_contig_name_for_my_sequence\n": 56}


def test_format_contig_nocut_notDuplicateName():
    """
    For a given contig, if we want to annotate it with prodigal, and do not cut,
    then we keep the same file (no need to split at 20 characters)
    However, we must check that contig names are all different.
    Add 2 contigs, to be sure the 'num' parameter is not increased.
    """
    cut = False
    pat = None
    cur_seq = "AACTGCTTTTTAAGCGCGCTCCTGCGNNNNNGGTTGTGTGGGCCCAGAGCGAGNCG"
    cur_seq2 = 'AACGTGGTCAGAGCGTG'
    cur_contig_name = ">my_contig_name_for_my_sequence"
    cur_contig_name2 = ">mycontigname"
    contig_sizes = {">mycontig": 155}
    num = 1

    assert gfunc.format_contig(cut, pat, cur_seq, cur_contig_name, "genome", contig_sizes, None,
                               num, logger=None) == 1
    assert gfunc.format_contig(cut, pat, cur_seq2, cur_contig_name2, "genome", contig_sizes, None,
                               num, logger=None) == 1

    assert contig_sizes == {">my_contig_name_for_my_sequence": 56,
                            ">mycontigname": 17,
                            ">mycontig": 155}


def test_format_contig_nocut_DuplicateName(caplog):
    """
    For a given contig, if we want to annotate it with prodigal, and do not cut, then we keep the same file. However, we must check that contig names are all different.
    Try to add a contig which name is already used, check that it prints the expected error,
    and returns -1
    """
    cut = False
    pat = None
    cur_seq = "AACTGCTTTTTAAGCGCGCTCCTGCGNNNNNGGTTGTGTGGGCCCAGAGCGAGNCG"
    cur_seq2 = 'AACGTGGTCAGAGCGTG'
    cur_contig_name = ">my_contig_name_for_my_sequence"
    cur_contig_name2 = ">my_contig2"
    contig_sizes = {">mycontig": 155, ">my_contig_name_for_my_sequence":45}
    num = 1

    # Try to add a contig already existing -> error log
    assert gfunc.format_contig(cut, pat, cur_seq, cur_contig_name, "genome", contig_sizes, None,
                               num, logger) == -1
    assert contig_sizes == {">my_contig_name_for_my_sequence": 45,
                            ">mycontig": 155}
    # Check logs
    caplog.set_level(logging.DEBUG)
    assert ("In genome genome, '>my_contig_name_for_my_sequence' contig name is used for "
            "several contigs. "
            "Please put different names for each contig. This genome will be "
            "ignored.") in caplog.text

    # Add a contig with new name. contig_sizes is completed
    assert gfunc.format_contig(cut, pat, cur_seq2, cur_contig_name2, "genome", contig_sizes, None,
                               num, logger) == 1
    assert contig_sizes == {">my_contig_name_for_my_sequence": 45,
                            ">my_contig2": 17,
                            ">mycontig": 155}


def test_get_outdir_prodigal_nocut():
  """
  When we use prodigal, and do not cut at each stretch of 5N, no need to create a modified
  sequence. So, check that get_output_dir returns an empty res_path (as we will not create
  any new sequence)
  """
  soft = "prodigal"
  genome = "genome1.fasta"
  cut = False
  pat = None
  gpath, grespath = gfunc.get_output_dir(soft, GEN_PATH, GENEPATH, genome, cut, pat)
  assert not grespath
  assert gpath == os.path.join(GEN_PATH, "genome1.fasta")


def test_get_outdir_prodigal_cut():
  """
  When using prodigal, and cut at each stretch of 4 'N', check that the output
  dir where modified sequence must be store is as expected (in tmp_dir, and adding "-split4N").
  """
  soft = "prodigal"
  genome = "genome1.fasta"
  cut = True
  pat = "NNNN+"
  gpath, grespath = gfunc.get_output_dir(soft, GEN_PATH, GENEPATH, genome, cut, pat)
  assert gpath == os.path.join(GEN_PATH, "genome1.fasta")
  assert grespath == os.path.join(GENEPATH, "genome1.fasta_prodigal-split4N.fna")


def test_get_outdir_prokka_nocut():
  """
  When using prokka, and don't cut sequence, check that the output
  dir where modified sequence must be store is as expected (in tmp dir, indicating
  "shorter_contigs" because we need to shorten the contig headers for prokka)
  """
  soft = "prokka"
  genome = "genome1.fasta"
  cut = False
  pat = None
  gpath, grespath = gfunc.get_output_dir(soft, GEN_PATH, GENEPATH, genome, cut, pat)
  assert gpath == os.path.join(GEN_PATH, "genome1.fasta")
  assert not grespath


def test_get_outdir_prokka_cut():
  """
  When using prokka, and cut at each stretch of 3 'N', check that the output
  dir where modified sequence must be store is as expected (in tmp_di, indicating "-split3N"
  because sequence will be cut)
  """
  soft = "prokka"
  genome = "genome1.fasta"
  cut = True
  pat = "NNN+"
  gpath, grespath = gfunc.get_output_dir(soft, GEN_PATH, GENEPATH, genome, cut, pat)
  assert gpath == os.path.join(GEN_PATH, "genome1.fasta")
  assert grespath == os.path.join(GENEPATH, "genome1.fasta_prokka-split3N.fna")


def test_get_outdir_no_input_seq():
    """
    Test that when the given genome name does not exist in the dbpath, it exists in tmp path.
    It means that it is a concatenation of 2 files of dbpath, which was saved in tmppath
    """
    soft = "prodigal"
    genome = "prokka_out_for_test-supHeader.faa"
    tmp_path = os.path.join("test", "data", "annotate", "test_files")
    cut = True
    pat = "NNNNNNN+"
    gpath, grespath = gfunc.get_output_dir(soft, GEN_PATH, tmp_path, genome, cut, pat)
    assert gpath == os.path.join(tmp_path, "prokka_out_for_test-supHeader.faa")
    assert grespath == os.path.join(tmp_path,
                                  "prokka_out_for_test-supHeader.faa_prodigal-split7N.fna")


def test_analyse1genome_nocut_prodigal():
    """
    Analyse the given genome, without cutting at stretches of N, will be annotated by prodigal:
    -> no new file created.
    Calculate its genome size, nb contigs and L90, and add it to the genomes dict, as well as
    the path to the genome file.
    """
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta"]
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"]}
    genome = gs[1]
    cut = False
    pat = None
    soft = "prodigal"
    assert gfunc.analyse_genome(genome, GEN_PATH, GENEPATH, cut, pat, genomes, soft, logger)

    # Check that information on analyzed genome are correct. And path to 'genome to annotate'
    # is the same as the path to the genome itself
    outf = os.path.join(GEN_PATH, "genome2.fasta")
    exp_genomes = {gs[0]: ["SAEN.1113"],
                   gs[1]: ["SAEN.1114", outf, outf, 67, 3, 3],
                   gs[2]: ["ESCO.0416"]}
    assert genomes == exp_genomes


def test_analyse1genome_cut_prodigal():
    '''
    Analyse the given genome, cutting at stretches of 5N, in order to annotate it
    Create new genome file in outdir, calculate genome size, nb contigs and L90, and add it
    to the genomes dict, as well as the path to the genome file.
    '''
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta"]
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"]}
    genome = gs[1]
    cut = True
    pat = "NNNNN+"
    soft = "prodigal"
    assert gfunc.analyse_genome(genome, GEN_PATH, GENEPATH, cut, pat, genomes, soft, logger)

    # Check that information on analyzed genome are correct. And path to 'genome to annotate'
    # is the same as the path to the genome itself
    initf = os.path.join(GEN_PATH, "genome2.fasta")  # initial genome path
    outf = os.path.join(GENEPATH, "genome2.fasta_prodigal-split5N.fna")  # path to geerated genome
    exp_out = os.path.join(EXP_DIR, "genome2-split5N.fna") # expected generated genome
    assert os.path.isfile(outf)
    assert tutil.compare_order_content(outf, exp_out)
    exp_genomes = {gs[0]: ["SAEN.1113"],
                   gs[1]: ["SAEN.1114", initf, outf, 55, 5, 4],
                   gs[2]: ["ESCO.0416"]}
    assert genomes == exp_genomes


def test_analyse1genome_cut_prokka():
    '''
    Analyse the given genome, cutting at stretches of 5N, in order to annotate it with prokka
    Create new genome file in outdir, with shortened contig names, calculate genome size,
    nb contigs and L90, and add it
    to the genomes dict, as well as the path to the genome file.
    '''
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta"]
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"]}
    genome = gs[1]
    cut = True
    pat = "NNNNN+"
    soft = "prokka"
    assert gfunc.analyse_genome(genome, GEN_PATH, GENEPATH, cut, pat, genomes, soft, logger)

    # Check that information on analyzed genome are correct. And path to 'genome to annotate'
    # is the same as the path to the genome itself
    initf = os.path.join(GEN_PATH, "genome2.fasta")  # initial genome path
    outf = os.path.join(GENEPATH, "genome2.fasta_prokka-split5N.fna")  # path to geerated genome
    exp_out = os.path.join(EXP_DIR, "genome2-split5N.fna") # expected generated genome
    assert os.path.isfile(outf)
    assert tutil.compare_order_content(outf, exp_out)
    exp_genomes = {gs[0]: ["SAEN.1113"],
                   gs[1]: ["SAEN.1114", initf, outf, 55, 5, 4],
                   gs[2]: ["ESCO.0416"]}
    assert genomes == exp_genomes


def test_analyse1genome_cut_same_names():
    """
    Analyse a genome. Its contig names all have the same first 20 characters. There is no
    stretch of at least 5N, so contigs are not split.
    New contig names should be uniq, and not all starting with '1_'!
    """
    genome = "genome_long_header.fst"
    genomes = {genome: ["SAEN.1015.0117"]}
    cut = True
    pat = "NNNNN+"
    assert gfunc.analyse_genome(genome, GEN_PATH, GENEPATH, cut, pat, genomes, "prokka", logger)
    in_f = os.path.join(GEN_PATH, genome)  # genome given
    out_f = os.path.join(GENEPATH, genome + "_prokka-split5N.fna")  # genome generated
    exp_f = os.path.join(EXP_DIR, "res_genome_short-long_header.fst") # expected genome generated
    exp_genomes = {genome: ["SAEN.1015.0117", in_f, out_f, 151, 3, 3]}
    assert genomes == exp_genomes
    assert tutil.compare_order_content(out_f, exp_f)


def test_analyse1genome_same_names_nocut(caplog):
    """
    Analyse a genome. 2 contigs have the same name, and we do not generate a new file (no cut)
    should return false with corresponding error message
    """
    caplog.set_level(logging.DEBUG)
    genome = "genome_2_identical_headers.fst"
    genomes = {genome: ["SAEN.1015.0117"]}
    cut = False
    pat = None
    assert not gfunc.analyse_genome(genome, GEN_PATH, GENEPATH, cut, pat, genomes,
                                    "prodigal", logger)
    assert ("In genome genome_2_identical_headers.fst, '>myheader' contig name is used for "
            "several contigs. Please put different names for "
            "each contig. This genome will be ignored") in caplog.text


def test_analyse1genome_same_last_name_nocut(caplog):
    """
    Analyse a genome. 2 contigs have the same name, and we do not generate a new file (no cut)
    should return false with corresponding error message
    """
    caplog.set_level(logging.DEBUG)
    genome = "genome_2_identical_headers-lastContig.fst"
    genomes = {genome: ["SAEN.1015.0117"]}
    cut = False
    pat = None
    assert not gfunc.analyse_genome(genome, GEN_PATH, GENEPATH, cut, pat, genomes,
                                    "prodigal", logger)
    assert ("In genome genome_2_identical_headers-lastContig.fst, '>myheader' contig name "
            "is used for several contigs. Please put different names for "
            "each contig. This genome will be ignored") in caplog.text


def test_analyse1genome_nofile(caplog):
    '''
    Test that when we ask to analyse a genome whose sequence file does not exist, it returns false
    with corresponding error message.
    '''
    caplog.set_level(logging.DEBUG)
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta"]
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"]}
    genome = "toto"
    cut = True
    pat = "NNNNN+"
    soft = "prodigal"
    assert not gfunc.analyse_genome(genome, GEN_PATH, GENEPATH, cut, pat, genomes, soft, logger)
    assert "The file toto does not exist"


def test_analyse1genome_empty(caplog):
    '''
    Test that when we ask to analyse a genome whose sequence file does not exist, it returns false
    with corresponding error message.
    '''
    caplog.set_level(logging.DEBUG)
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta"]
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"]}
    genome = "genome1.fasta"
    # create empty genome file
    open(os.path.join(GENEPATH, genome), "w").close()
    cut = True
    pat = "NNNNN+"
    soft = "prodigal"
    assert not gfunc.analyse_genome(genome, GENEPATH, GENEPATH, cut, pat, genomes, soft, logger)
    assert ("Your file test/data/annotate/generated_by_unit-tests/genome1.fasta does not contain "
            "any gene. Please check that you really gave a fasta sequence file") in caplog.text
    gres = os.path.join(GENEPATH, "toto.fna_prodigal-split5N.fna")
    assert not os.path.isfile(gres)


def test_analyse_all_genomes_nocut(caplog):
    """
    Analyze all given genomes: don't cut at stretches of N, but look at their sequence
    file, to calculate L90, genome size and nb contigs. Add this information, as well as the
    path to the genomic sequence, to the genomes dict.
    """
    caplog.set_level(logging.DEBUG)
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta"]
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"]}
    nbn = 0
    # Run analysis
    gfunc.analyse_all_genomes(genomes, GEN_PATH, GENEPATH, nbn, "prokka", logger, quiet=False)
    # construct expected results
    gpaths = [os.path.join(GEN_PATH, gname) for gname in gs]
    exp_genomes = {gs[0]: ["SAEN.1113", gpaths[0], gpaths[0], 51, 4, 2],
                   gs[1]: ["SAEN.1114", gpaths[1], gpaths[1], 67, 3, 3],
                   gs[2]: ["ESCO.0416", gpaths[2], gpaths[2], 70, 4, 1]}
    assert exp_genomes == genomes
    assert ("Calculating genome size, number of contigs, L90") in caplog.text


def test_analyse_all_genomes_binary(caplog):
    """
    Analyze all given genomes: don't cut at stretches of N, but look at their sequence
    file, to calculate L90, genome size and nb contigs. Add this information, as well as the
    path to the genomic sequence, to the genomes dict.
    1 file is a binary file: write warning message and remove it from analysis.
    """
    caplog.set_level(logging.DEBUG)
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta", "genome.fna.bin"]
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"],
               gs[3]: ["BIN.1234"]}
    nbn = 0
    # Run analysis
    gfunc.analyse_all_genomes(genomes, GEN_PATH, GENEPATH, nbn, "prokka", logger, quiet=False)
    # construct expected results
    gpaths = [os.path.join(GEN_PATH, gname) for gname in gs]
    exp_genomes = {gs[0]: ["SAEN.1113", gpaths[0], gpaths[0], 51, 4, 2],
                   gs[1]: ["SAEN.1114", gpaths[1], gpaths[1], 67, 3, 3],
                   gs[2]: ["ESCO.0416", gpaths[2], gpaths[2], 70, 4, 1]}
    assert exp_genomes == genomes
    assert ("Calculating genome size, number of contigs, L90") in caplog.text
    assert ("'genome.fna.bin' does not seem to be a fasta file. It "
            "will be ignored.") in caplog.text


def test_analyse_all_genomes_cut(caplog):
    """
    Analyze all given genomes: cut at stretches of 3N, and look at their sequence
    file, to calculate L90, genome size and nb contigs. Add this information, as well as the
    path to the genomic sequence, to the genomes dict.
    """
    caplog.set_level(logging.DEBUG)
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta"]
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"]}
    nbn = 3
    # Run analysis
    gfunc.analyse_all_genomes(genomes, GEN_PATH, GENEPATH, nbn, "prokka", logger, quiet=False)
    # construct expected results
    gpaths = [os.path.join(GEN_PATH, gname) for gname in gs]
    opaths = [os.path.join(GENEPATH, gname + "_prokka-split3N.fna") for gname in gs]
    exp_genomes = {gs[0]: ["SAEN.1113", gpaths[0], opaths[0], 51, 4, 2],
                   gs[1]: ["SAEN.1114", gpaths[1], opaths[1], 51, 6, 5],
                   gs[2]: ["ESCO.0416", gpaths[2], opaths[2], 70, 4, 1]}
    assert exp_genomes == genomes
    assert ("Cutting genomes at each time there are at least 3 'N' in a row, "
            "and then, calculating genome size, number of contigs and L90.") in caplog.text


def test_analyse_all_genomes_nocut_empty(caplog):
    """
    Analyze all given genomes: don't cut at stretches of N, but look at their sequence
    file, to calculate L90, genome size and nb contigs. Add this information, as well as the
    path to the genomic sequence, to the genomes dict. 1 genome is empty -> should be removed
    """
    caplog.set_level(logging.DEBUG)
    gs = ["genome1.fasta", "genome2.fasta", "empty.fasta", "genome3.fasta"]
    empty_genome = os.path.join(GEN_PATH, gs[2])
    # Add an empty genome to the original database
    open(empty_genome, "w").close()
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"],
               gs[3]: ["ESCO.0123"]}
    nbn = 0
    # Run analysis
    gfunc.analyse_all_genomes(genomes, GEN_PATH, GENEPATH, nbn, "prokka", logger, quiet=True)
    # construct expected results
    gpaths = [os.path.join(GEN_PATH, gname) for gname in gs]
    exp_genomes = {gs[0]: ["SAEN.1113", gpaths[0], gpaths[0], 51, 4, 2],
                   gs[1]: ["SAEN.1114", gpaths[1], gpaths[1], 67, 3, 3],
                   gs[3]: ["ESCO.0123", gpaths[3], gpaths[3], 70, 4, 1]}
    assert exp_genomes == genomes
    assert ("Calculating genome size, number of contigs, L90") in caplog.text
    assert ("Your file test/data/annotate/genomes/empty.fasta "
            "does not contain any gene. Please check that you really gave a "
            "fasta sequence file") in caplog.text

    # remove the empty genome
    os.remove(empty_genome)


def test_analyse_all_genomes_cut_empty(caplog):
    """
    Analyze all given genomes: cut at stretches of 3N, and look at their sequence
    file, to calculate L90, genome size and nb contigs. Add this information, as well as the
    path to the genomic sequence, to the genomes dict. 1 genome is empty -> should be removed
    """
    caplog.set_level(logging.DEBUG)
    gs = ["genome1.fasta", "genome2.fasta", "empty.fasta", "genome3.fasta"]
    empty_genome = os.path.join(GEN_PATH, gs[2])
    # Add an empty genome to the original database
    open(empty_genome, "w").close()
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"],
               gs[3]: ["ESCO.0123"]}
    nbn = 3
    # Run analysis
    gfunc.analyse_all_genomes(genomes, GEN_PATH, GENEPATH, nbn, "prokka", logger, quiet=True)
    # construct expected results
    gpaths = [os.path.join(GEN_PATH, gname) for gname in gs]
    opaths = [os.path.join(GENEPATH, gname + "_prokka-split3N.fna") for gname in gs]
    exp_genomes = {gs[0]: ["SAEN.1113", gpaths[0], opaths[0], 51, 4, 2],
                   gs[1]: ["SAEN.1114", gpaths[1], opaths[1], 51, 6, 5],
                   gs[3]: ["ESCO.0123", gpaths[3], opaths[3], 70, 4, 1]}
    assert exp_genomes == genomes
    assert ("Cutting genomes at each time there are at least 3 'N' in a row, "
            "and then, calculating genome size, number of contigs and L90.") in caplog.text
    assert ("Your file test/data/annotate/genomes/empty.fasta "
            "does not contain any gene. Please check that you really gave a "
            "fasta sequence file") in caplog.text

    # remove the empty genome
    os.remove(empty_genome)


def test_analyse_all_genomes_noseq(caplog):
    """
    Analyze all given genomes: no given sequence file exists
    -> Exits with error message
    """
    caplog.set_level(logging.DEBUG)
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta", "genome4.fasta"]
    genomes = {gs[0]: ["SAEN.1113"],
               gs[1]: ["SAEN.1114"],
               gs[2]: ["ESCO.0416"],
               gs[3]: ["ESCO.0123"]}
    nbn = 3
    # Run analysis
    with pytest.raises(SystemExit):
        gfunc.analyse_all_genomes(genomes, "toto", GENEPATH, nbn, "prokka", logger, quiet=True)
    assert ("No genome was found in the database folder toto. See logfile "
            "for more information.") in caplog.text


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance=6, backend="agg")
def test_dist_l90():
    """
    For created L90 graph, check that calculated L90 values are as expected,
    and graph is also as expected
    """
    listfile_base = "test_plot_dist"
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta", "genome4.fasta",
          "genome5.fasta", "genome6.fasta", "genome7.fasta"]
    genomes = {gs[0]: ["SAEN.1113", 'orig_path',  'annot_path', 51, 2, 2],
               gs[1]: ["SAEN.1114", 'orig_path', 'annot_path', 67, 15, 13],
               gs[2]: ["ESCO.0416", 'orig_path', 'annot_path', 70, 15, 11],
               gs[3]: ["ESCO.0216", 'orig_path', 'annot_path', 114, 17, 11],
               gs[4]: ["SAEN.1115", 'orig_path', 'annot_path', 106, 17, 12],
               gs[5]: ["ESCO.0216", 'orig_path', 'annot_path', 116, 60, 50],
               gs[6]: ["SAEN.1115", 'orig_path', 'annot_path', 137, 20, 12]}
    l90 = 13
    nbconts = 19
    l90, _, dist, _ = gfunc.plot_distributions(genomes, GENEPATH, listfile_base, l90, nbconts)
    outfile1 = os.path.join(GENEPATH, "QC_L90-" + listfile_base + ".png")
    outfile2 = os.path.join(GENEPATH, "QC_nb-contigs-" + listfile_base + ".png")
    # Check that png file was created
    assert os.path.isfile(outfile1)
    assert os.path.isfile(outfile2)
    # Check values calculated for l90
    assert set(l90) == {2, 13, 11, 11, 12, 50, 12}
    # Check that output plot is as expected
    return dist



@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance=6, backend="agg")
def test_dist_nbcont():
    """
    For created L90 graph, check that calculated L90 values are as expected,
    and graph is also as expected
    """
    listfile_base = "test_plot_dist"
    gs = ["genome1.fasta", "genome2.fasta", "genome3.fasta", "genome4.fasta",
          "genome5.fasta", "genome6.fasta", "genome7.fasta"]
    genomes = {gs[0]: ["SAEN.1113", 'orig_path',  'annot_path', 51, 2, 2],
               gs[1]: ["SAEN.1114", 'orig_path', 'annot_path', 67, 15, 13],
               gs[2]: ["ESCO.0416", 'orig_path', 'annot_path', 70, 15, 11],
               gs[3]: ["ESCO.0216", 'orig_path', 'annot_path', 114, 17, 11],
               gs[4]: ["SAEN.1115", 'orig_path', 'annot_path', 106, 17, 12],
               gs[5]: ["ESCO.0216", 'orig_path', 'annot_path', 116, 60, 50],
               gs[6]: ["SAEN.1115", 'orig_path', 'annot_path', 137, 20, 12]}
    l90 = 13
    nbconts = 19
    _, nbcont, _, dist = gfunc.plot_distributions(genomes, GENEPATH, listfile_base, l90, nbconts)
    outfile1 = os.path.join(GENEPATH, "QC_L90-" + listfile_base + ".png")
    outfile2 = os.path.join(GENEPATH, "QC_nb-contigs-" + listfile_base + ".png")
    # Check that png file was created
    assert os.path.isfile(outfile1)
    assert os.path.isfile(outfile2)
    # Check values calculated for nbconts
    assert set(nbcont) == {2, 15, 15, 17, 17, 60, 20}
    # Check that output plot is as expected
    return dist
